@@ -1,31 +1,31 @@
 /*******************************************************************************
  *
- * Copyright (c) 2000-2003 Intel Corporation 
- * All rights reserved. 
- * Copyright (c) 2012 France Telecom All rights reserved. 
+ * Copyright (c) 2000-2003 Intel Corporation
+ * All rights reserved.
+ * Copyright (c) 2012 France Telecom All rights reserved.
  *
- * Redistribution and use in source and binary forms, with or without 
- * modification, are permitted provided that the following conditions are met: 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
  *
- * - Redistributions of source code must retain the above copyright notice, 
- * this list of conditions and the following disclaimer. 
- * - Redistributions in binary form must reproduce the above copyright notice, 
- * this list of conditions and the following disclaimer in the documentation 
- * and/or other materials provided with the distribution. 
- * - Neither name of Intel Corporation nor the names of its contributors 
- * may be used to endorse or promote products derived from this software 
+ * - Redistributions of source code must retain the above copyright notice,
+ * this list of conditions and the following disclaimer.
+ * - Redistributions in binary form must reproduce the above copyright notice,
+ * this list of conditions and the following disclaimer in the documentation
+ * and/or other materials provided with the distribution.
+ * - Neither name of Intel Corporation nor the names of its contributors
+ * may be used to endorse or promote products derived from this software
  * without specific prior written permission.
- * 
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
- * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR 
- * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
- * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
- * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
- * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR
+ * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
  * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  ******************************************************************************/
@@ -38,30 +38,26 @@
 
 #include <curl/curl.h>
 
-#include <string>
+#include <algorithm>
 #include <map>
 #include <sstream>
-#include <algorithm>
+#include <string>
 
 #include "gena.h"
-#include "httputils.h"
+#include "genut.h"
 #include "statcodes.h"
 #include "upnpapi.h"
-#include "upnp_timeout.h"
-#include "genut.h"
-#include "TimerThread.h"
-#include "gena_sids.h"
-#include "netif.h"
+#include "uri.h"
 
 #ifdef USE_EXPAT
-#include "expatmm.hxx"
+#include "expatmm.h"
 #define XMLPARSERTP inputRefXMLParser
 #else
 #include "picoxml.h"
 #define XMLPARSERTP PicoXMLParser
 #endif
 
-#define UPNP_NOPE (Upnp_LogLevel)(UPNP_ALL+1)
+#define UPNP_NOPE static_cast<Upnp_LogLevel>(UPNP_ALL+1)
 
 extern TimerThread *gTimerThread;
 
@@ -96,7 +92,18 @@ static void clientCancelRenew(ClientSubscription *sub)
     }
 }
 
-struct upnp_timeout_data_subscribe : public upnp_timeout_data {
+class AutoRenewSubscriptionJobWorker : public JobWorker {
+public:
+    explicit AutoRenewSubscriptionJobWorker(
+        int h, std::string s, int e, const std::string& url, int t) {
+        handle = h;
+        sub.Sid = std::move(s);
+        sub.ErrCode = e;
+        upnp_strlcpy(sub.PublisherUrl, url, NAME_SIZE);
+        sub.TimeOut = t;
+    }
+    void work() override;
+    int handle;
     struct Upnp_Event_Subscribe sub;
 };
 
@@ -104,49 +111,44 @@ struct upnp_timeout_data_subscribe : public upnp_timeout_data {
  * \brief This is a thread function to send the renewal just before the
  * subscription times out.
  */
-static void *thread_autorenewsubscription(
-    /*! [in] Thread data(upnp_timeout *) needed to send the renewal. */
-    void *input)
+void AutoRenewSubscriptionJobWorker::work()
 {
-    auto event = static_cast<upnp_timeout *>(input);
-    auto sub_struct = &(dynamic_cast<upnp_timeout_data_subscribe*>(event->Event)->sub);
     int send_callback = 0;
     Upnp_EventType eventType = UPNP_EVENT_AUTORENEWAL_FAILED;
 
-    if (AUTO_RENEW_TIME == 0) {
-        // We are compile-time configured for no auto-renewal.
-        UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__, "GENA SUB EXPIRED\n");
-        sub_struct->ErrCode = UPNP_E_SUCCESS;
+#if AUTO_RENEW_TIME == 0
+    // We are compile-time configured for no auto-renewal.
+    UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__, "GENA SUB EXPIRED\n");
+    sub.ErrCode = UPNP_E_SUCCESS;
+    send_callback = 1;
+    eventType = UPNP_EVENT_SUBSCRIPTION_EXPIRED;
+#else
+    UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__, "GENA AUTO RENEW\n");
+    int timeout = sub.TimeOut;
+    int errCode = genaRenewSubscription(handle, sub.Sid, &timeout);
+    sub.ErrCode = errCode;
+    sub.TimeOut = timeout;
+    if (errCode != UPNP_E_SUCCESS &&
+        errCode != UPNP_E_INVALID_SID &&
+        errCode != UPNP_E_INVALID_HANDLE) {
         send_callback = 1;
-        eventType = UPNP_EVENT_SUBSCRIPTION_EXPIRED;
-    } else {
-        UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__, "GENA AUTO RENEW\n");
-        int timeout = sub_struct->TimeOut;
-        int errCode = genaRenewSubscription(event->handle, sub_struct->Sid, &timeout);
-        sub_struct->ErrCode = errCode;
-        sub_struct->TimeOut = timeout;
-        if (errCode != UPNP_E_SUCCESS &&
-            errCode != GENA_E_BAD_SID &&
-            errCode != GENA_E_BAD_HANDLE) {
-            send_callback = 1;
-        }
     }
+#endif
 
     if (send_callback) {
-        HandleReadLock();
+        Upnp_FunPtr callback_fun;
         struct Handle_Info *handle_info;
-        if (GetHandleInfo(event->handle, &handle_info) != HND_CLIENT) {
-            HandleUnlock();
-            free_upnp_timeout(event);
-            return nullptr;
+        void *cookie;
+        {
+            HANDLELOCK();
+            if (GetHandleInfo(handle, &handle_info) != HND_CLIENT) {
+                return;
+            }
+            callback_fun = handle_info->Callback;
+            cookie = handle_info->Cookie;
         }
-
-        /* make callback */
-        Upnp_FunPtr callback_fun = handle_info->Callback;
-        HandleUnlock();
-        callback_fun(eventType, sub_struct, handle_info->Cookie);
+        callback_fun(eventType, &sub, cookie);
     }
-    return nullptr;
 }
 
 
@@ -165,34 +167,20 @@ static int ScheduleGenaAutoRenew(
     ClientSubscription *sub)
 {
     if (TimeOut == UPNP_INFINITE) {
-        return GENA_SUCCESS;
+        return UPNP_E_SUCCESS;
     }
 
-    auto RenewEventStruct = new upnp_timeout_data_subscribe;
-    auto RenewEvent =  new upnp_timeout;
-
-    /* schedule expire event */
-    RenewEventStruct->sub.Sid = sub->SID;
-    RenewEventStruct->sub.ErrCode = UPNP_E_SUCCESS;
-    upnp_strlcpy(RenewEventStruct->sub.PublisherUrl, sub->eventURL, NAME_SIZE);
-    RenewEventStruct->sub.TimeOut = TimeOut;
-
-    RenewEvent->handle = client_handle;
-    RenewEvent->Event = RenewEventStruct;
-
     /* Schedule the job */
+    auto worker = std::make_unique<AutoRenewSubscriptionJobWorker>(
+        client_handle, sub->SID, UPNP_E_SUCCESS, sub->eventURL, TimeOut);
     int return_code = gTimerThread->schedule(
         TimerThread::SHORT_TERM, TimerThread::REL_SEC, TimeOut - AUTO_RENEW_TIME,
-        &(RenewEvent->eventId), thread_autorenewsubscription, RenewEvent,
-        reinterpret_cast<ThreadPool::free_routine>(free_upnp_timeout));
+        &sub->renewEventId, std::move(worker));
 
     if (return_code != UPNP_E_SUCCESS) {
-        free_upnp_timeout(RenewEvent);
         return return_code;
     }
-
-    sub->renewEventId = RenewEvent->eventId;
-    return GENA_SUCCESS;
+    return UPNP_E_SUCCESS;
 }
 
 
@@ -205,7 +193,8 @@ static int gena_unsubscribe(
     /*! [in] Event URL of the service. */
     const std::string& url,
     /*! [in] The subcription ID. */
-    const std::string& sid)
+    const std::string& sid,
+    int timeoutms)
 {
     int return_code;
     uri_type dest_url;
@@ -226,7 +215,7 @@ static int gena_unsubscribe(
     curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, "UNSUBSCRIBE");
     std::string surl =  uri_asurlstr(dest_url);
     curl_easy_setopt(easy, CURLOPT_URL, surl.c_str());
-    curl_easy_setopt(easy, CURLOPT_TIMEOUT, HTTP_DEFAULT_TIMEOUT);
+    curl_easy_setopt(easy, CURLOPT_TIMEOUT_MS, timeoutms);
 
     struct curl_slist *list = nullptr;
     list = curl_slist_append(list, (std::string("SID: ") + sid).c_str());
@@ -257,7 +246,7 @@ static int gena_unsubscribe(
 }
 
 // localaddr is already in inet_ntop-provided dot or ipv6 format
-static std::string myCallbackUrl(NetIF::IPAddr& netaddr)
+static std::string myCallbackUrl(const NetIF::IPAddr& netaddr)
 {
     std::ostringstream oss;
     oss << "http://";
@@ -273,8 +262,7 @@ static std::string myCallbackUrl(NetIF::IPAddr& netaddr)
 }
 
 struct CurlGuard {
-    CURL *htalk{nullptr};
-    struct curl_slist *hlist{nullptr};
+    CurlGuard(CURL* t, curl_slist* l) : htalk(t), hlist(l) {}
     ~CurlGuard() {
         if (htalk) {
             curl_easy_cleanup(htalk);
@@ -283,6 +271,10 @@ struct CurlGuard {
             curl_slist_free_all(hlist);
         }
     }
+    CurlGuard(const CurlGuard&) = delete;
+    CurlGuard& operator=(const CurlGuard&) = delete;
+    CURL *htalk;
+    struct curl_slist *hlist;
 };
 
 /*!
@@ -299,7 +291,8 @@ static int gena_subscribe(
      * For first time subscription, this must be empty. */
     const std::string& renewal_sid,
     /*! [out] SID returned by the subscription or renew msg. */
-    std::string *sid)
+    std::string *sid,
+    int timeoutms)
 {
     int local_timeout = CP_MINIMUM_SUBSCRIPTION_TIME;
 
@@ -317,7 +310,7 @@ static int gena_subscribe(
     } else {
         timostr << *timeout;
     }
-    
+
     /* parse url */
     uri_type dest_url;
     int return_code = http_FixStrUrl(url, &dest_url);
@@ -325,43 +318,47 @@ static int gena_subscribe(
         return return_code;
     }
     std::string urlforcurl = uri_asurlstr(dest_url);
-    NetIF::IPAddr destaddr(
-        reinterpret_cast<struct sockaddr*>(&dest_url.hostport.IPaddress));
+    NetIF::IPAddr destaddr(reinterpret_cast<struct sockaddr*>(&dest_url.hostport.IPaddress));
+
+    // Determine a suitable address for the callback. We choose one on the interface for the
+    // destination address. Another possible approach would be to actually connect to the URL and
+    // use getsockname(), which would let the routing code do the main job, at the cost of a
+    // supplementary connection.
     NetIF::IPAddr myaddr;
     const NetIF::Interface *ifp =
-        NetIF::Interfaces::theInterfaces()->interfaceForAddress(destaddr,myaddr);
+        NetIF::Interfaces::theInterfaces()->interfaceForAddress(destaddr, g_netifs, myaddr);
     if (nullptr == ifp) {
         UpnpPrintf(UPNP_ERROR, GENA, __FILE__, __LINE__,
                    "Could not find the interface for the destination address\n");
         return UPNP_E_SOCKET_CONNECT;
     }
-       
-    CurlGuard hdls;
+
     std::map<std::string, std::string> http_headers;
     char curlerrormessage[CURL_ERROR_SIZE];
-    
-    hdls.htalk = curl_easy_init();
+
+    auto hdls = CurlGuard(curl_easy_init(), nullptr);
     curl_easy_setopt(hdls.htalk, CURLOPT_ERRORBUFFER, curlerrormessage);
     curl_easy_setopt(hdls.htalk, CURLOPT_WRITEFUNCTION, write_callback_null_curl);
     curl_easy_setopt(hdls.htalk, CURLOPT_CUSTOMREQUEST, "SUBSCRIBE");
     curl_easy_setopt(hdls.htalk, CURLOPT_URL, urlforcurl.c_str());
-    curl_easy_setopt(hdls.htalk, CURLOPT_TIMEOUT, HTTP_DEFAULT_TIMEOUT);
+    curl_easy_setopt(hdls.htalk, CURLOPT_TIMEOUT_MS, timeoutms);
     curl_easy_setopt(hdls.htalk, CURLOPT_HEADERFUNCTION, header_callback_curl);
     curl_easy_setopt(hdls.htalk, CURLOPT_HEADERDATA, &http_headers);
+    std::string descript;
     if (renewal_sid.empty()) {
         std::string cbheader{"CALLBACK: <"};
         cbheader += myCallbackUrl(myaddr) + "/>";
         hdls.hlist = curl_slist_append(hdls.hlist, cbheader.c_str());
         hdls.hlist = curl_slist_append(hdls.hlist, "NT: upnp:event");
-        UpnpPrintf(UPNP_ALL,GENA,__FILE__,__LINE__,
-                   "gena_subscribe(init.): url [%s] cb [%s] timeout [%s]\n",
-                   urlforcurl.c_str(), myCallbackUrl(myaddr).c_str(), timostr.str().c_str());
+        descript = std::string("(init) ") + "url [" + urlforcurl  + "] cb [" +
+            myCallbackUrl(myaddr) + "] timeout [" + timostr.str() + "]";
+        UpnpPrintf(UPNP_ALL,GENA,__FILE__,__LINE__, "gena_subscribe: %s\n", descript.c_str());
     } else {
         hdls.hlist = curl_slist_append(
             hdls.hlist, (std::string("SID: ") + renewal_sid).c_str());
-        UpnpPrintf(UPNP_ALL,GENA,__FILE__,__LINE__,
-                   "gena_subscribe(renew): SID [%s] url [%s] timeout [%s]\n",
-                   renewal_sid.c_str(), urlforcurl.c_str(), timostr.str().c_str());
+        descript = std::string("(renew) ") + "url [" + urlforcurl  + "] SID [" +
+            renewal_sid + "] timeout [" + timostr.str() + "]";
+        UpnpPrintf(UPNP_ALL,GENA,__FILE__,__LINE__, "gena_subscribe: %s\n", descript.c_str());
     }
     hdls.hlist = curl_slist_append(
         hdls.hlist, (std::string("TIMEOUT: Second-") + timostr.str()).c_str());
@@ -372,15 +369,16 @@ static int gena_subscribe(
     CURLcode curlcode = curl_easy_perform(hdls.htalk);
     if (curlcode != CURLE_OK) {
         /* We may want to detail things here, depending on the curl error */
-        UpnpPrintf(UPNP_ERROR,GENA,__FILE__,__LINE__, "CURL ERROR MESSAGE %s\n", curlerrormessage);
+        UpnpPrintf(UPNP_ERROR,GENA,__FILE__,__LINE__,
+                   "gena_subscribe: %s: CURL ERROR MESSAGE %s\n", descript.c_str(),curlerrormessage);
         return UPNP_E_SOCKET_CONNECT;
     }
 
     long http_status;
     curl_easy_getinfo (hdls.htalk, CURLINFO_RESPONSE_CODE, &http_status);
     if (http_status != HTTP_OK) {
-        UpnpPrintf(UPNP_DEBUG,GENA,__FILE__,__LINE__, "gena_subscribe: HTTP status %d\n",
-                   int(http_status));
+        UpnpPrintf(UPNP_DEBUG,GENA,__FILE__,__LINE__,
+                   "gena_subscribe: %s: HTTP status %d\n", descript.c_str(), int(http_status));
         return UPNP_E_SUBSCRIBE_UNACCEPTED;
     }
 
@@ -402,75 +400,73 @@ static int gena_subscribe(
     *sid = itsid->second;
     UpnpPrintf(UPNP_ALL,GENA,__FILE__,__LINE__, "gena_subscribe ok: SID [%s] timeout %d\n",
                itsid->second.c_str(), *timeout);
-    
+
     return UPNP_E_SUCCESS;
 }
 
 
 int genaUnregisterClient(UpnpClient_Handle client_handle)
 {
-    int return_code = UPNP_E_SUCCESS;
-    struct Handle_Info *handle_info = nullptr;
-
+    struct Handle_Info *handle_info;
+    int timeoutms;
+    ClientSubscription sub_copy;
+    
     while (true) {
-        HandleLock();
+        {
+            HANDLELOCK();
 
-        if (GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
-            HandleUnlock();
-            return GENA_E_BAD_HANDLE;
+            if (GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
+                return UPNP_E_INVALID_HANDLE;
+            }
+            if (handle_info->ClientSubList.empty()) {
+                break;
+            }
+            sub_copy = handle_info->ClientSubList.front();
+            handle_info->ClientSubList.remove_if(
+                [sub_copy] (const ClientSubscription& e) {return e.SID == sub_copy.SID;});
+
+            timeoutms = handle_info->SubsOpsTimeoutMS;
         }
-        if (handle_info->ClientSubList.empty()) {
-            return_code = UPNP_E_SUCCESS;
-            break;
-        }
-        ClientSubscription sub_copy = handle_info->ClientSubList.front();
-        handle_info->ClientSubList.remove_if(
-            [sub_copy] (const ClientSubscription& e) {return e.SID == sub_copy.SID;});
 
-        HandleUnlock();
-
-        gena_unsubscribe(sub_copy.eventURL, sub_copy.SID);
+        gena_unsubscribe(sub_copy.eventURL, sub_copy.SID, timeoutms);
         clientCancelRenew(&sub_copy);
     }
 
-    handle_info->ClientSubList.clear();
-    HandleUnlock();
-
-    return return_code;
+    return UPNP_E_SUCCESS;
 }
 
 
 int genaUnSubscribe(UpnpClient_Handle client_handle, const std::string& in_sid)
 {
-    /* validate handle and sid */
-    HandleLock();
     struct Handle_Info *handle_info;
-    if (GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
-        HandleUnlock();
-        return GENA_E_BAD_HANDLE;
+    int timeoutms;
+    ClientSubscription sub_copy;
+    {
+        /* validate handle and sid */
+        HANDLELOCK();
+        if (GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
+            return UPNP_E_INVALID_HANDLE;
+        }
+        auto sub = std::find_if(handle_info->ClientSubList.begin(), handle_info->ClientSubList.end(),
+                                [in_sid](const ClientSubscription& e){return e.SID == in_sid;});
+        if (handle_info->ClientSubList.end() == sub) {
+            return UPNP_E_INVALID_SID;
+        }
+        timeoutms = handle_info->SubsOpsTimeoutMS;
+        sub_copy = *sub;
     }
-    auto sub = std::find_if(handle_info->ClientSubList.begin(), handle_info->ClientSubList.end(),
-                            [in_sid](const ClientSubscription& e){return e.SID == in_sid;});
-    if (handle_info->ClientSubList.end() == sub) {
-        HandleUnlock();
-        return GENA_E_BAD_SID;
-    }
-    auto sub_copy = *sub;
-    HandleUnlock();
 
-    gena_unsubscribe(sub_copy.eventURL, sub_copy.SID);
+    gena_unsubscribe(sub_copy.eventURL, sub_copy.SID, timeoutms);
     clientCancelRenew(&sub_copy);
 
-    HandleLock();
+    HANDLELOCK();
     if (GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
-        HandleUnlock();
-        return GENA_E_BAD_HANDLE;
+        return UPNP_E_INVALID_HANDLE;
     }
     handle_info->ClientSubList.remove_if(
         [in_sid] (const ClientSubscription& e) {return e.SID == in_sid;});
-    HandleUnlock();
 
-    return GENA_SUCCESS;
+    return UPNP_E_SUCCESS;
 }
 
 
@@ -480,54 +476,51 @@ int genaSubscribe(
     int *TimeOut,
     std::string *out_sid)
 {
-    int return_code = GENA_SUCCESS;
-    ClientSubscription newSubscription;
+    int return_code = UPNP_E_SUCCESS;
     std::string SID;
     std::string EventURL;
     struct Handle_Info *handle_info;
-
+    int timeoutms;
+    
     out_sid->clear();
 
-    HandleReadLock();
-    /* validate handle */
-    if (GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
-        return_code = GENA_E_BAD_HANDLE;
-        SubscribeLock();
-        goto error_handler;
+    {
+        HANDLELOCK();
+        /* validate handle */
+        if (GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
+            return UPNP_E_INVALID_HANDLE;
+        }
+        timeoutms = handle_info->SubsOpsTimeoutMS;
     }
-    HandleUnlock();
 
     /* subscribe */
     SubscribeLock();
-    return_code = gena_subscribe(PublisherURL, TimeOut, std::string(), &SID);
-    HandleLock();
+    return_code = gena_subscribe(PublisherURL, TimeOut, std::string(), &SID, timeoutms);
     if (return_code != UPNP_E_SUCCESS) {
         UpnpPrintf(UPNP_ERROR, GENA, __FILE__, __LINE__,
                    "genaSubscribe: subscribe error, return %d\n", return_code);
         goto error_handler;
     }
 
-    if(GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
-        return_code = GENA_E_BAD_HANDLE;
-        goto error_handler;
+    {
+        HANDLELOCK();
+        if(GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
+            return_code = UPNP_E_INVALID_HANDLE;
+            goto error_handler;
+        }
+
+        /* create event url */
+        EventURL = PublisherURL;
+        out_sid->assign(SID);
+        handle_info->ClientSubList.emplace_front(-1, std::move(SID), std::move(EventURL));
+
+        /* schedule expiration event */
+        return_code = ScheduleGenaAutoRenew(client_handle, *TimeOut,
+                                            &handle_info->ClientSubList.front());
     }
 
-    /* create event url */
-    EventURL = PublisherURL;
-    newSubscription.renewEventId = -1;
-    newSubscription.SID = SID;
-    out_sid->assign(SID);
-    newSubscription.eventURL = EventURL;
-    handle_info->ClientSubList.push_front(newSubscription);
-
-    /* schedule expiration event */
-    return_code = ScheduleGenaAutoRenew(client_handle, *TimeOut,
-                                        &handle_info->ClientSubList.front());
-
 error_handler:
-    HandleUnlock();
     SubscribeUnlock();
-
     return return_code;
 }
 
@@ -537,38 +530,38 @@ int genaRenewSubscription(
     const std::string& in_sid,
     int *TimeOut)
 {
-    HandleLock();
-
-    /* validate handle and sid */
     struct Handle_Info *handle_info;
-    if (GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
-        HandleUnlock();
-        return GENA_E_BAD_HANDLE;
+    int timeoutms;
+    ClientSubscription sub_copy;
+    {
+        HANDLELOCK();
+
+        /* validate handle and sid */
+        if (GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
+            return UPNP_E_INVALID_HANDLE;
+        }
+
+        auto sub = std::find_if(handle_info->ClientSubList.begin(), handle_info->ClientSubList.end(),
+                                [in_sid](const ClientSubscription& e){return e.SID == in_sid;});
+        if (handle_info->ClientSubList.end() == sub) {
+            return UPNP_E_INVALID_SID;
+        }
+        timeoutms = handle_info->SubsOpsTimeoutMS;
+
+        /* remove old events */
+        gTimerThread->remove(sub->renewEventId);
+
+        sub->renewEventId = -1;
+        sub_copy = *sub;
     }
 
-    auto sub = std::find_if(handle_info->ClientSubList.begin(), handle_info->ClientSubList.end(),
-                            [in_sid](const ClientSubscription& e){return e.SID == in_sid;});
-    if (handle_info->ClientSubList.end() == sub) {
-        HandleUnlock();
-        return GENA_E_BAD_SID;
-    }
-
-    /* remove old events */
-    gTimerThread->remove(sub->renewEventId);
-
-    sub->renewEventId = -1;
-    ClientSubscription sub_copy = *sub;
-
-    HandleUnlock();
-
     std::string SID;
-    int return_code = gena_subscribe(sub_copy.eventURL, TimeOut, sub_copy.SID, &SID);
+    int return_code = gena_subscribe(sub_copy.eventURL, TimeOut, sub_copy.SID, &SID, timeoutms);
 
-    HandleLock();
+    HANDLELOCK();
 
     if (GetHandleInfo(client_handle, &handle_info) != HND_CLIENT) {
-        HandleUnlock();
-        return GENA_E_BAD_HANDLE;
+        return UPNP_E_INVALID_HANDLE;
     }
 
     if (return_code != UPNP_E_SUCCESS) {
@@ -576,17 +569,15 @@ int genaRenewSubscription(
         handle_info->ClientSubList.remove_if(
             [in_sid] (const ClientSubscription& e) {return e.SID == in_sid;});
         clientCancelRenew(&sub_copy);
-        HandleUnlock();
         return return_code;
     }
 
     /* get subscription */
-    sub = std::find_if(handle_info->ClientSubList.begin(), handle_info->ClientSubList.end(),
-                       [in_sid](const ClientSubscription& e){return e.SID == in_sid;});
+    auto sub = std::find_if(handle_info->ClientSubList.begin(), handle_info->ClientSubList.end(),
+                            [in_sid](const ClientSubscription& e){return e.SID == in_sid;});
     if (handle_info->ClientSubList.end() == sub) {
         clientCancelRenew(&sub_copy);
-        HandleUnlock();
-        return GENA_E_BAD_SID;
+        return UPNP_E_INVALID_SID;
     }
 
     /* Remember SID */
@@ -594,13 +585,12 @@ int genaRenewSubscription(
 
     /* start renew subscription timer */
     return_code = ScheduleGenaAutoRenew(client_handle, *TimeOut, &*sub);
-    if (return_code != GENA_SUCCESS) {
+    if (return_code != UPNP_E_SUCCESS) {
         handle_info->ClientSubList.remove_if(
             [sub] (const ClientSubscription& e) {return e.SID == sub->SID;});
     }
     clientCancelRenew(&sub_copy);
 
-    HandleUnlock();
     return return_code;
 }
 
@@ -609,7 +599,7 @@ public:
     UPnPPropertysetParser(
         // XML to be parsed
         const std::string& input,
-        // Output data 
+        // Output data
         std::unordered_map<std::string, std::string>& propd)
         : XMLPARSERTP(input),  propdata(propd) {
     }
@@ -640,7 +630,7 @@ private:
 void gena_process_notification_event(MHDTransaction *mhdt)
 {
     UpnpPrintf(UPNP_ALL, GENA, __FILE__, __LINE__, "gena_process_notification_event\n");
-    
+
     auto itsid = mhdt->headers.find("sid");
     /* get SID */
     if (itsid == mhdt->headers.end()) {
@@ -698,14 +688,14 @@ void gena_process_notification_event(MHDTransaction *mhdt)
                    mhdt->postdata.c_str());
         return;
     }
-    HandleLock();
+    globalHndLock.lock();
 
     /* get client info */
     struct Handle_Info *handle_info;
     UpnpClient_Handle client_handle;
     if (GetClientHandleInfo(&client_handle, &handle_info) != HND_CLIENT) {
         http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
-        HandleUnlock();
+        globalHndLock.unlock();
         return;
     }
 
@@ -719,19 +709,19 @@ void gena_process_notification_event(MHDTransaction *mhdt)
             /*   (if we are in the middle) */
             /* this is to avoid mistakenly rejecting the first event if we  */
             /*   receive it before the subscription response */
-            HandleUnlock();
+            globalHndLock.unlock();
 
             /* try and get Subscription Lock  */
             /*   (in case we are in the process of subscribing) */
             SubscribeLock();
 
             /* get HandleLock again */
-            HandleLock();
+            globalHndLock.lock();
 
             if (GetClientHandleInfo(&client_handle,&handle_info) != HND_CLIENT) {
                 http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
                 SubscribeUnlock();
-                HandleUnlock();
+                globalHndLock.unlock();
                 return;
             }
 
@@ -741,7 +731,7 @@ void gena_process_notification_event(MHDTransaction *mhdt)
             if (handle_info->ClientSubList.end() == subscription) {
                 http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
                 SubscribeUnlock();
-                HandleUnlock();
+                globalHndLock.unlock();
                 return;
             }
 
@@ -750,9 +740,9 @@ void gena_process_notification_event(MHDTransaction *mhdt)
             UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__,
                        "gena_process_notification_event: could not find subscription "
                        "but event key not 0 (%d)\n", eventKey);
-            
+
             http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
-            HandleUnlock();
+            globalHndLock.unlock();
             return;
         }
     }
@@ -770,7 +760,7 @@ void gena_process_notification_event(MHDTransaction *mhdt)
     auto callback = handle_info->Callback;
     auto cookie = handle_info->Cookie;
 
-    HandleUnlock();
+    globalHndLock.unlock();
 
     /* make callback with event struct */
     /* In future, should find a way of mainting */
