@@ -1,32 +1,32 @@
 /*******************************************************************************
  *
- * Copyright (c) 2000-2003 Intel Corporation 
- * All rights reserved. 
- * Copyright (c) 2012 France Telecom All rights reserved. 
- * Copyright (c) 2020 J.F. Dockes <jf@dockes.org>
+ * Copyright (c) 2000-2003 Intel Corporation
+ * All rights reserved.
+ * Copyright (c) 2012 France Telecom All rights reserved.
+ * Copyright (c) 2020-2024 J.F. Dockes <jf@dockes.org>
  *
- * Redistribution and use in source and binary forms, with or without 
- * modification, are permitted provided that the following conditions are met: 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
  *
- * - Redistributions of source code must retain the above copyright notice, 
- * this list of conditions and the following disclaimer. 
- * - Redistributions in binary form must reproduce the above copyright notice, 
- * this list of conditions and the following disclaimer in the documentation 
- * and/or other materials provided with the distribution. 
- * - Neither name of Intel Corporation nor the names of its contributors 
- * may be used to endorse or promote products derived from this software 
+ * - Redistributions of source code must retain the above copyright notice,
+ * this list of conditions and the following disclaimer.
+ * - Redistributions in binary form must reproduce the above copyright notice,
+ * this list of conditions and the following disclaimer in the documentation
+ * and/or other materials provided with the distribution.
+ * - Neither name of Intel Corporation nor the names of its contributors
+ * may be used to endorse or promote products derived from this software
  * without specific prior written permission.
- * 
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
- * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR 
- * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
- * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
- * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
- * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR
+ * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
  * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  ******************************************************************************/
@@ -36,19 +36,16 @@
 #if EXCLUDE_GENA == 0
 #ifdef INCLUDE_DEVICE_APIS
 
-#include <sstream>
 #include <iostream>
+#include <sstream>
 
 #include <curl/curl.h>
 
-#include <cassert>
-
 #include "gena.h"
-#include "httputils.h"
-#include "statcodes.h"
-#include "upnpapi.h"
 #include "gena_sids.h"
 #include "genut.h"
+#include "statcodes.h"
+#include "upnpapi.h"
 #include "uri.h"
 
 static constexpr auto XML_PROPERTYSET_HEADER =
@@ -57,31 +54,26 @@ static constexpr auto XML_PROPERTYSET_HEADER =
 /*!
  * \brief Unregisters a device.
  *
- * \return UPNP_E_SUCCESS on success, GENA_E_BAD_HANDLE on failure.
+ * \return UPNP_E_SUCCESS on success, UPNP_E_INVALID_HANDLE on failure.
  */
 int genaUnregisterDevice(UpnpDevice_Handle device_handle)
 {
-    int ret = 0;
     struct Handle_Info *handle_info;
 
-    HandleLock();
+    HANDLELOCK();
     if (GetHandleInfo(device_handle, &handle_info) != HND_DEVICE) {
         UpnpPrintf(UPNP_CRITICAL, GENA, __FILE__, __LINE__,
             "genaUnregisterDevice: BAD Handle: %d\n", device_handle);
-        ret = GENA_E_BAD_HANDLE;
-    } else {
-        freeServiceTable(&handle_info->ServiceTable);
-        ret = UPNP_E_SUCCESS;
+        return UPNP_E_INVALID_HANDLE;
     }
-    HandleUnlock();
-
-    return ret;
+    clearServiceTable(handle_info->serviceTable);
+    return UPNP_E_SUCCESS;
 }
 
 /*!
  * \brief Generates XML property set for notifications.
  *
- * \return UPNP_E_SUCCESS if successful else returns GENA_E_BAD_HANDLE.
+ * \return UPNP_E_SUCCESS if successful else returns UPNP_E_INVALID_HANDLE.
  *
  * \note The XML_VERSION comment is NOT sent due to interoperability issues
  *     with other UPnP vendors.
@@ -116,13 +108,13 @@ static int GeneratePropertySet(
  *
  * NOTIFY http request is sent and the reply is processed.
  *
- * \return GENA_SUCCESS if the event was delivered, otherwise returns the
+ * \return UPNP_E_SUCCESS if the event was delivered, otherwise returns the
  *     appropriate error code.
  *
  * The only code which has specific processing is
- * GENA_E_NOTIFY_UNACCEPTED_REMOVE_SUB which results in clearing the 
+ * GENA_E_NOTIFY_UNACCEPTED_REMOVE_SUB which results in clearing the
  * subscription. This can only happen if the HTTP transaction is
- * network-successful but has an HTTP status of 412, which results in 
+ * network-successful but has an HTTP status of 412, which results in
  * clearing the subscription (else, subscriptions are only removed
  * when they time-out).
  * The previous version had more detailed error codes, but all other
@@ -130,10 +122,9 @@ static int GeneratePropertySet(
  */
 static int genaNotify(const std::string& propertySet, const subscription *sub)
 {
-    std::string mid_msg;
     int return_code = -1;
-
     long http_code = 0;
+
     /* send a notify to each url until one goes thru */
     for (const auto& url : sub->DeliveryURLs) {
         CURL *easy = curl_easy_init();
@@ -146,17 +137,16 @@ static int genaNotify(const std::string& propertySet, const subscription *sub)
                          long(GENA_NOTIFICATION_SENDING_TIMEOUT +
                               GENA_NOTIFICATION_ANSWERING_TIMEOUT)/2);
         curl_easy_setopt(easy, CURLOPT_POST, long(1));
-        curl_easy_setopt(easy, CURLOPT_POSTFIELDS, propertySet.c_str()); 
-        curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, "NOTIFY"); 
+        curl_easy_setopt(easy, CURLOPT_POSTFIELDS, propertySet.c_str());
+        curl_easy_setopt(easy, CURLOPT_CUSTOMREQUEST, "NOTIFY");
 
         struct curl_slist *list = nullptr;
         list = curl_slist_append(list, "NT: upnp:event");
         list = curl_slist_append(list, "NTS: upnp:propchange");
         list = curl_slist_append(list,(std::string("SID: ") + sub->sid).c_str());
-        char buff[100];
-        snprintf(buff, 100, "%d", sub->ToSendEventKey);
+        auto buff = std::to_string(sub->ToSendEventKey);
         list = curl_slist_append(list, (std::string("SEQ: ") + buff).c_str());
-        
+
         list = curl_slist_append(list, "Accept:");
         list = curl_slist_append(list, "Expect:");
         list = curl_slist_append(list, R"(Content-Type: text/xml; charset="utf-8")");
@@ -183,13 +173,13 @@ static int genaNotify(const std::string& propertySet, const subscription *sub)
 
     if (return_code == UPNP_E_SUCCESS) {
         if (http_code == HTTP_OK) {
-            return_code = GENA_SUCCESS;
+            return_code = UPNP_E_SUCCESS;
         } else {
             if (http_code == HTTP_PRECONDITION_FAILED)
                 /*Invalid SID gets removed */
                 return_code = GENA_E_NOTIFY_UNACCEPTED_REMOVE_SUB;
             else
-                return_code = GENA_E_NOTIFY_UNACCEPTED;
+                return_code = UPNP_E_NOTIFY_UNACCEPTED;
         }
     }
     return return_code;
@@ -202,6 +192,10 @@ static int genaNotify(const std::string& propertySet, const subscription *sub)
    though (e.g. with shared_ptr for the propertySet string. We don't
    typically have thousands of subscribed CPs... */
 struct Notification {
+    Notification(std::string sid, std::string udn, std::string pset, Upnp_SID uSID,
+                 time_t ct, UpnpDevice_Handle dh)
+        : device_handle(dh), UDN(std::move(udn)), servId(std::move(sid)), sid(std::move(uSID)),
+          propertySet(std::move(pset)), ctime(ct) {}
     UpnpDevice_Handle device_handle; //
     std::string UDN;                 // Device
     std::string servId;  // Service
@@ -210,12 +204,19 @@ struct Notification {
     time_t ctime;        // Age
 };
 
-/* free_func set when creating the ThreadPool job. Always called by
-   the ThreadPool when deleting the job after done or error. */
-static void free_notify_struct(Notification *input)
-{
-    delete input;
-}
+class GenaNotifyJobWorker : public JobWorker {
+public:
+    explicit GenaNotifyJobWorker(std::shared_ptr<Notification> in)
+        : m_input(std::move(in)) {}
+    void work() override;
+    // This is actually shared with the outgoing list head.  Note: 2024-02: as far as I understand,
+    // the only reason to keep a shared pointer to the active notification at the head of the
+    // subscription outgoing queue is as an indicator that we don't need to kickstart the
+    // queue. Else when adding an event and seeing an empty queue, we would not know if the event
+    // should be queued or directly sent to the Thread Pool. This could probably be replaced by a
+    // flag, which would allow to replace the shared_ptr with unique_ptr.
+    std::shared_ptr<Notification> m_input;
+};
 
 /*!
  * \brief Thread job to Notify a control point.
@@ -223,69 +224,56 @@ static void free_notify_struct(Notification *input)
  * It validates and copies the subscription so that the lock can be
  * released during the actual network transfer (done by genaNotify())
  */
-static void *thread_genanotify(void *input)
+void GenaNotifyJobWorker::work()
 {
-    auto in = static_cast<Notification *>(input);
     subscription *sub;
     service_info *service;
     subscription sub_copy;
     int return_code;
     struct Handle_Info *handle_info;
 
-    /* This should be a HandleLock and not a HandleReadLock otherwise if there
-     * is a lot of notifications, then multiple threads will acquire a read
-     * lock and the thread which sends the notification will be blocked forever
-     * on the HandleLock at the end of this function. */
-    /*HandleReadLock(); */
-    HandleLock();
-    /* validate context */
+    {
+        HANDLELOCK();
+        /* validate context */
+        if (GetHandleInfo(m_input->device_handle, &handle_info) != HND_DEVICE) {
+            return;
+        }
 
-    if (GetHandleInfo(in->device_handle, &handle_info) != HND_DEVICE) {
-        HandleUnlock();
-        return nullptr;
+        if (!(service = FindServiceId(handle_info->serviceTable, m_input->servId, m_input->UDN)) ||
+            !service->active ||
+            !(sub = GetSubscriptionSID(m_input->sid, service)) ||
+            copy_subscription(sub, &sub_copy) != UPNP_E_SUCCESS) {
+            return;
+        }
     }
 
-    if (!(service = FindServiceId(&handle_info->ServiceTable,
-                                  in->servId, in->UDN)) ||
-        !service->active ||
-        !(sub = GetSubscriptionSID(in->sid, service)) ||
-        copy_subscription(sub, &sub_copy) != UPNP_E_SUCCESS) {
-        HandleUnlock();
-        return nullptr;
-    }
-
-    HandleUnlock();
-
     /* send the notify */
-    return_code = genaNotify(in->propertySet, &sub_copy);
-    HandleLock();
-    if (GetHandleInfo(in->device_handle, &handle_info) != HND_DEVICE) {
-        HandleUnlock();
-        return nullptr;
+    return_code = genaNotify(m_input->propertySet, &sub_copy);
+
+    HANDLELOCK();
+    if (GetHandleInfo(m_input->device_handle, &handle_info) != HND_DEVICE) {
+        return;
     }
     /* validate context */
-    if (!(service = FindServiceId(&handle_info->ServiceTable,
-                                  in->servId, in->UDN)) ||
+    if (!(service = FindServiceId(handle_info->serviceTable, m_input->servId, m_input->UDN)) ||
         !service->active ||
-        !(sub = GetSubscriptionSID(in->sid, service))) {
-        HandleUnlock();
-        return nullptr;
+        !(sub = GetSubscriptionSID(m_input->sid, service))) {
+        return;
     }
     sub->ToSendEventKey++;
     if (sub->ToSendEventKey < 0)
         /* wrap to 1 for overflow */
         sub->ToSendEventKey = 1;
 
-    /* Remove head of event queue. Do not delete it, the ThreadJob free_func
-       will do it */ 
+    /* Remove head of event queue. */
     if (!sub->outgoing.empty()) {
         sub->outgoing.pop_front();
     }
     /* Possibly activate next */
     if (!sub->outgoing.empty()) {
         auto notif = sub->outgoing.begin();
-        gSendThreadPool.addJob(thread_genanotify, *notif,
-                               reinterpret_cast<ThreadPool::free_routine>(free_notify_struct));
+        auto worker = std::make_unique<GenaNotifyJobWorker>(*notif);
+        gSendThreadPool.addJob(std::move(worker));
     }
 
     // No idea why we do this after sending one more event. Was the
@@ -294,10 +282,7 @@ static void *thread_genanotify(void *input)
     // the first Notif then, because it's the only case where it's not
     // managed by a ThreadPool Job (potentially creating a mem leak).
     if (return_code == GENA_E_NOTIFY_UNACCEPTED_REMOVE_SUB)
-        RemoveSubscriptionSID(in->sid, service);
-
-    HandleUnlock();
-    return nullptr;
+        RemoveSubscriptionSID(m_input->sid, service);
 }
 
 
@@ -306,13 +291,11 @@ int genaInitNotifyXML(
     char *UDN,
     char *servId,
     const std::string& propertySet,
-    const Upnp_SID sid)
+    const Upnp_SID& sid)
 {
-    int ret = GENA_SUCCESS;
+    int ret = UPNP_E_SUCCESS;
     int line = 0;
-
-    Notification *thread_struct = nullptr;
-
+    std::shared_ptr<Notification> thread_struct;
     subscription *sub = nullptr;
     service_info *service = nullptr;
     struct Handle_Info *handle_info;
@@ -320,61 +303,47 @@ int genaInitNotifyXML(
     UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__,
                "genaInitNotifyXML: props: %s\n", propertySet.c_str());
 
-    HandleLock();
+    HANDLELOCK();
 
     if (GetHandleInfo(device_handle, &handle_info) != HND_DEVICE) {
         line = __LINE__;
-        ret = GENA_E_BAD_HANDLE;
+        ret = UPNP_E_INVALID_HANDLE;
         goto ExitFunction;
     }
 
-    service = FindServiceId(&handle_info->ServiceTable, servId, UDN);
+    service = FindServiceId(handle_info->serviceTable, servId, UDN);
     if (service == nullptr) {
         line = __LINE__;
-        ret = GENA_E_BAD_SERVICE;
+        ret = UPNP_E_INVALID_SERVICE;
         goto ExitFunction;
     }
 
     sub = GetSubscriptionSID(sid, service);
     if (sub == nullptr || sub->active) {
         line = __LINE__;
-        ret = GENA_E_BAD_SID;
+        ret = UPNP_E_INVALID_SID;
         goto ExitFunction;
     }
     sub->active = 1;
 
     /* schedule thread for initial notification */
-    thread_struct = new Notification;
-    if (thread_struct == nullptr) {
-        line = __LINE__;
-        ret = UPNP_E_OUTOF_MEMORY;
-        goto ExitFunction;
+    thread_struct = std::make_shared<Notification>(
+        servId, UDN, propertySet, sid, time(nullptr), device_handle);
+    {
+        auto worker = std::make_unique<GenaNotifyJobWorker>(thread_struct);
+        ret = gSendThreadPool.addJob(std::move(worker));
     }
-    thread_struct->servId = servId;
-    thread_struct->UDN = UDN;
-    thread_struct->propertySet = propertySet;
-    thread_struct->sid = sid;
-    thread_struct->ctime = time(nullptr);
-    thread_struct->device_handle = device_handle;
-
-    ret = gSendThreadPool.addJob(thread_genanotify, thread_struct,
-                                 reinterpret_cast<ThreadPool::free_routine>(free_notify_struct));
     if (ret != 0) {
         line = __LINE__;
         ret = UPNP_E_OUTOF_MEMORY;
     } else {
         line = __LINE__;
         sub->outgoing.push_back(thread_struct);
-        ret = GENA_SUCCESS;
+        ret = UPNP_E_SUCCESS;
     }
 
 ExitFunction:
-    if (ret != GENA_SUCCESS) {
-        delete thread_struct;
-    }
-    HandleUnlock();
-    UpnpPrintf(UPNP_ALL, GENA, __FILE__, line,
-               "genaInitNotifyCommon: ret %d\n", ret);
+    UpnpPrintf(UPNP_ALL, GENA, __FILE__, line, "genaInitNotifyCommon: ret %d\n", ret);
     return ret;
 }
 
@@ -385,18 +354,17 @@ int genaInitNotifyVars(
     char **VarNames,
     char **VarValues,
     int var_count,
-    const Upnp_SID sid)
+    const Upnp_SID& sid)
 {
-    int ret = GENA_SUCCESS;
+    int ret = UPNP_E_SUCCESS;
     int line = 0;
     std::string propertySet;
 
-    UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__,
-               "genaInitNotifyVars varcnt %d\n", var_count);
+    UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__, "genaInitNotifyVars varcnt %d\n", var_count);
 
     if (var_count <= 0) {
         line = __LINE__;
-        ret = GENA_SUCCESS;
+        ret = UPNP_E_SUCCESS;
         goto ExitFunction;
     }
 
@@ -413,21 +381,6 @@ ExitFunction:
     return ret;
 }
 
-void freeSubscriptionQueuedEvents(subscription *sub)
-{
-    /* The first event is discarded without deleting the Notification:
-       there is a mirror ThreadPool entry for this one, and the
-       completion event will take care of it. Other entries must be
-       fully cleaned-up here */
-    auto it = sub->outgoing.begin();
-    if (it != sub->outgoing.end())
-        it++;
-    while (it != sub->outgoing.end()) {
-        delete *it;
-        it = sub->outgoing.erase(it);
-    }
-}
-
 /*
  * This gets called before queuing a new event, with the handLock held.
  * - The list size can never go over MAX_SUBSCRIPTION_QUEUED_EVENTS so we
@@ -436,7 +389,7 @@ void freeSubscriptionQueuedEvents(subscription *sub)
  * non-active: any but the head of queue, which is already copied to
  * the thread pool
  */
-static void maybeDiscardEvents(std::list<Notification*>& outgoing)
+static void maybeDiscardEvents(std::list<std::shared_ptr<Notification>>& outgoing)
 {
     time_t now = time(nullptr);
 
@@ -445,10 +398,8 @@ static void maybeDiscardEvents(std::list<Notification*>& outgoing)
     if (it != outgoing.end())
         it++;
     while (it != outgoing.end()) {
-        Notification *ntsp = *it;
         if (outgoing.size() > unsigned(g_UpnpSdkEQMaxLen) ||
-            now - ntsp->ctime > g_UpnpSdkEQMaxAge) {
-            delete ntsp;
+            now - (*it)->ctime > g_UpnpSdkEQMaxAge) {
             it = outgoing.erase(it);
         } else {
             /* If the list is smaller than the max and the oldest
@@ -459,60 +410,44 @@ static void maybeDiscardEvents(std::list<Notification*>& outgoing)
 }
 
 int genaNotifyAllXML(
-    UpnpDevice_Handle device_handle,
-    char *UDN, char *servId, const std::string& propertySet)
+    UpnpDevice_Handle device_handle, char *UDN, char *servId, const std::string& propertySet)
 {
-    int ret = GENA_SUCCESS;
+    int ret = UPNP_E_SUCCESS;
     int line = 0;
     std::list<subscription>::iterator finger;
-    
-    Notification *thread_struct = nullptr;
-
+    std::shared_ptr<Notification> thread_struct;
     service_info *service = nullptr;
     struct Handle_Info *handle_info;
 
     UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__,
                "genaNotifyAllXML: props: %s\n", propertySet.c_str());
 
-    HandleLock();
+    HANDLELOCK();
 
     if (GetHandleInfo(device_handle, &handle_info) != HND_DEVICE) {
         line = __LINE__;
-        ret = GENA_E_BAD_HANDLE;
-        goto ExitFunction;
-    } 
-
-    service = FindServiceId(&handle_info->ServiceTable, servId, UDN);
-    if (service == nullptr) {
-        line = __LINE__;
-        ret = GENA_E_BAD_SERVICE;
+        ret = UPNP_E_INVALID_HANDLE;
         goto ExitFunction;
     }
-    
+
+    service = FindServiceId(handle_info->serviceTable, servId, UDN);
+    if (service == nullptr) {
+        line = __LINE__;
+        ret = UPNP_E_INVALID_SERVICE;
+        goto ExitFunction;
+    }
+
     finger = GetFirstSubscription(service);
     while (finger != service->subscriptionList.end()) {
-        thread_struct = new Notification;
-        if (thread_struct == nullptr) {
-            line = __LINE__;
-            ret = UPNP_E_OUTOF_MEMORY;
-            break;
-        }
-        thread_struct->UDN = UDN;
-        thread_struct->servId = servId;
-        thread_struct->propertySet = propertySet;
-        thread_struct->ctime = time(nullptr);
-        thread_struct->device_handle = device_handle;
-        thread_struct->sid = finger->sid;
-
+        thread_struct = std::make_shared<Notification>(
+            servId, UDN, propertySet, finger->sid, time(nullptr), device_handle);
         maybeDiscardEvents(finger->outgoing);
         finger->outgoing.push_back(thread_struct);
 
-        /* If there is only one element on the list (which we just
-           added), need to kickstart the threadpool */
+        /* If there is only one element on the list (just added), kickstart the threadpool */
         if (finger->outgoing.size() == 1) {
-            ret = gSendThreadPool.addJob(
-                thread_genanotify, thread_struct,
-                reinterpret_cast<ThreadPool::free_routine>(free_notify_struct));
+            auto worker = std::make_unique<GenaNotifyJobWorker>(thread_struct);
+            ret = gSendThreadPool.addJob(std::move(worker));
             if (ret != 0) {
                 line = __LINE__;
                 if (ret == EOUTOFMEM) {
@@ -526,11 +461,7 @@ int genaNotifyAllXML(
     }
 
 ExitFunction:
-    HandleUnlock();
-
-    UpnpPrintf(UPNP_ALL, GENA, __FILE__, line,
-               "genaNotifyAllCommon: ret = %d\n", ret);
-
+    UpnpPrintf(UPNP_ALL, GENA, __FILE__, line, "genaNotifyAllCommon: ret = %d\n", ret);
     return ret;
 }
 
@@ -542,7 +473,7 @@ int genaNotifyAll(
     char **VarValues,
     int var_count)
 {
-    int ret = GENA_SUCCESS;
+    int ret = UPNP_E_SUCCESS;
     int line = 0;
 
     UpnpPrintf(UPNP_ALL, GENA, __FILE__, __LINE__, "genaNotifyAll\n");
@@ -600,7 +531,7 @@ static int respond_ok(MHDTransaction *mhdt, int time_out, subscription *sub,
 // one of our supported addresses, not necessarily the one this
 // particular client connected to.
 static bool callStrangerCheck(
-    const std::string& surl, uri_type& temp, const NetIF::Interface *clnetif, NetIF::IPAddr& claddr)
+    const std::string& surl, uri_type& temp, const NetIF::Interface *clnetif, const NetIF::IPAddr& claddr)
 {
     NetIF::IPAddr subsaddr(reinterpret_cast<struct sockaddr*>(&temp.hostport.IPaddress));
     if (!subsaddr.ok()) {
@@ -648,12 +579,12 @@ static int create_url_list(
     out->clear();
 
     // Get information for the client address
-    NetIF::IPAddr claddr(reinterpret_cast<struct sockaddr*>(mhdt->client_address));
+    NetIF::IPAddr claddr(reinterpret_cast<struct sockaddr*>(&mhdt->client_address));
     if (!claddr.ok()) {
         UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
                    "create_url_list: can't determine client addr\n");
         return UPNP_E_INVALID_INTERFACE;
-    }        
+    }
     UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__,
                "create_url_list: client address: %s\n",claddr.straddr().c_str());
     NetIF::IPAddr hostaddr;
@@ -665,7 +596,7 @@ static int create_url_list(
         return UPNP_E_INVALID_INTERFACE;
     }
 
-    std::string::size_type openpos = 0;
+    std::string::size_type openpos;
     std::string::size_type closepos = 0;
     for (;;) {
         if ((openpos = ulist.find('<', closepos)) == std::string::npos) {
@@ -709,14 +640,13 @@ static int create_url_list(
 
         UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__,
                    "create_url_list: cb url ok:  %s\n", qsurl.c_str());
-        out->push_back(qsurl);
+        out->push_back(std::move(qsurl));
     }
     return UPNP_E_SUCCESS;
 }
 
 void gena_process_subscription_request(MHDTransaction *mhdt)
 {
-    struct Upnp_Subscription_Request request_struct = {};
     int return_code = 1;
     int time_out = 1801;
     service_info *service;
@@ -724,7 +654,6 @@ void gena_process_subscription_request(MHDTransaction *mhdt)
     void *cookie;
     Upnp_FunPtr callback_fun;
     UpnpDevice_Handle device_handle;
-    std::list<subscription>::iterator sub;
 
     UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
                "Subscription Request Received:\n");
@@ -750,100 +679,92 @@ void gena_process_subscription_request(MHDTransaction *mhdt)
     UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__,
                "SubscriptionRequest for event URL path: %s\n",mhdt->url.c_str());
 
-    HandleLock();
-
-    if (GetDeviceHandleInfoForPath(
-            mhdt->url, &device_handle, &handle_info, &service) != HND_DEVICE) {
-        http_SendStatusResponse(mhdt, HTTP_INTERNAL_SERVER_ERROR);
-        HandleUnlock();
-        return;
-    }
-
-    if (service == nullptr || !service->active) {
-        http_SendStatusResponse(mhdt, HTTP_NOT_FOUND);
-        HandleUnlock();
-        return;
-    }
-
-    UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
-               "Subscription Request: current subscriptions count %d max %d\n",
-               service->TotalSubscriptions, handle_info->MaxSubscriptions);
-
-    /* too many subscriptions */
-    if (handle_info->MaxSubscriptions != -1 &&
-        service->TotalSubscriptions >= handle_info->MaxSubscriptions) {
-        http_SendStatusResponse(mhdt, HTTP_INTERNAL_SERVER_ERROR);
-        HandleUnlock();
-        return;
-    }
-
-    std::vector<std::string> tmpUrls;
-    /* check for valid callbacks */
+    Upnp_Subscription_Request request_struct;
     {
-        auto itcb = mhdt->headers.find("callback");
-        if (itcb == mhdt->headers.end()) {
-            http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
-            HandleUnlock();
+        HANDLELOCK();
+
+        if (GetDeviceHandleInfoForPath(
+                mhdt->url, &device_handle, &handle_info, &service) != HND_DEVICE) {
+            http_SendStatusResponse(mhdt, HTTP_INTERNAL_SERVER_ERROR);
             return;
         }
-        return_code = create_url_list(mhdt, itcb->second, &tmpUrls);
-        if (return_code != UPNP_E_SUCCESS) {
-            http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
-            HandleUnlock();
+
+        if (service == nullptr || !service->active) {
+            http_SendStatusResponse(mhdt, HTTP_NOT_FOUND);
             return;
         }
-    }
 
-    /* generate new subscription */
-    sub = service->subscriptionList.emplace(service->subscriptionList.end());
-    if (sub == service->subscriptionList.end()) {
-        http_SendStatusResponse(mhdt, HTTP_INTERNAL_SERVER_ERROR);
-        HandleUnlock();
-        return;
-    }
+        UpnpPrintf(UPNP_INFO, GENA, __FILE__, __LINE__,
+                   "Subscription Request: current subscriptions count %d max %d\n",
+                   service->TotalSubscriptions, handle_info->MaxSubscriptions);
 
-    sub->DeliveryURLs = tmpUrls;
-    /* set the timeout */
-    if (!timeout_header_value(mhdt->headers, &time_out)) {
-        time_out = GENA_DEFAULT_TIMEOUT;
-    }
-    
-    /* replace infinite timeout with max timeout, if possible */
-    if (handle_info->MaxSubscriptionTimeOut != -1) {
-        if (time_out == -1 || time_out > handle_info->MaxSubscriptionTimeOut) {
-            time_out = handle_info->MaxSubscriptionTimeOut;
+        /* too many subscriptions */
+        if (handle_info->MaxSubscriptions != -1 &&
+            service->TotalSubscriptions >= handle_info->MaxSubscriptions) {
+            http_SendStatusResponse(mhdt, HTTP_INTERNAL_SERVER_ERROR);
+            return;
         }
+
+        std::vector<std::string> tmpUrls;
+        /* check for valid callbacks */
+        {
+            auto itcb = mhdt->headers.find("callback");
+            if (itcb == mhdt->headers.end()) {
+                http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
+                return;
+            }
+            return_code = create_url_list(mhdt, itcb->second, &tmpUrls);
+            if (return_code != UPNP_E_SUCCESS) {
+                http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
+                return;
+            }
+        }
+
+        /* generate new subscription */
+        auto sub = service->subscriptionList.emplace(service->subscriptionList.end());
+        if (sub == service->subscriptionList.end()) {
+            http_SendStatusResponse(mhdt, HTTP_INTERNAL_SERVER_ERROR);
+            return;
+        }
+
+        sub->DeliveryURLs = tmpUrls;
+        /* set the timeout */
+        if (!timeout_header_value(mhdt->headers, &time_out)) {
+            time_out = GENA_DEFAULT_TIMEOUT;
+        }
+
+        /* replace infinite timeout with max timeout, if possible */
+        if (handle_info->MaxSubscriptionTimeOut != -1) {
+            if (time_out == -1 || time_out > handle_info->MaxSubscriptionTimeOut) {
+                time_out = handle_info->MaxSubscriptionTimeOut;
+            }
+        }
+        if (time_out >= 0) {
+            sub->expireTime = time(nullptr) + time_out;
+        } else {
+            /* infinite time */
+            sub->expireTime = 0;
+        }
+
+        /* generate SID */
+        sub->sid = std::string("uuid:") + gena_sid_uuid();
+
+        /* respond OK */
+        if (respond_ok(mhdt, time_out, &(*sub), handle_info->productversion) != UPNP_E_SUCCESS) {
+            service->subscriptionList.pop_back();
+            return;
+        }
+        service->TotalSubscriptions++;
+        UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__, "Subscription Request granted\n");
+
+        /* finally generate callback for init table dump */
+        request_struct = Upnp_Subscription_Request{
+            service->serviceId.c_str(), service->UDN.c_str(), sub->sid};
+
+        /* copy callback */
+        callback_fun = handle_info->Callback;
+        cookie = handle_info->Cookie;
     }
-    if (time_out >= 0) {
-        sub->expireTime = time(nullptr) + time_out;
-    } else {
-        /* infinite time */
-        sub->expireTime = 0;
-    }
-
-    /* generate SID */
-    sub->sid = std::string("uuid:") + gena_sid_uuid();
-
-    /* respond OK */
-    if (respond_ok(mhdt, time_out, &(*sub), handle_info->productversion) != UPNP_E_SUCCESS) {
-        service->subscriptionList.pop_back();
-        HandleUnlock();
-        return;
-    }
-    service->TotalSubscriptions++;
-    UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__,
-               "Subscription Request granted\n");
-
-    /* finally generate callback for init table dump */
-    request_struct.ServiceId = service->serviceId.c_str();
-    request_struct.UDN = service->UDN.c_str();
-    request_struct.Sid = sub->sid;
-
-    /* copy callback */
-    callback_fun = handle_info->Callback;
-    cookie = handle_info->Cookie;
-
-    HandleUnlock();
 
     /* make call back with request struct */
     /* in the future should find a way of mainting that the handle */
@@ -858,7 +779,6 @@ void gena_process_subscription_renewal_request(MHDTransaction *mhdt)
     service_info *service;
     struct Handle_Info *handle_info;
     UpnpDevice_Handle device_handle;
-    std::string event_url_path;
 
     /* if a CALLBACK or NT header is present, then it is an error */
     if (mhdt->headers.find("callback") != mhdt->headers.end() ||
@@ -876,12 +796,11 @@ void gena_process_subscription_renewal_request(MHDTransaction *mhdt)
 
     Upnp_SID sid = itsid->second;
 
-    HandleLock();
-    
+    HANDLELOCK();
+
     if (GetDeviceHandleInfoForPath(
             mhdt->url, &device_handle, &handle_info, &service) != HND_DEVICE ) {
         http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
-        HandleUnlock();
         return;
     }
 
@@ -889,7 +808,6 @@ void gena_process_subscription_renewal_request(MHDTransaction *mhdt)
     if(service == nullptr || !service->active ||
        ((sub = GetSubscriptionSID( sid, service )) == nullptr)) {
         http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
-        HandleUnlock();
         return;
     }
 
@@ -902,7 +820,6 @@ void gena_process_subscription_renewal_request(MHDTransaction *mhdt)
        service->TotalSubscriptions > handle_info->MaxSubscriptions ) {
         http_SendStatusResponse(mhdt, HTTP_INTERNAL_SERVER_ERROR);
         RemoveSubscriptionSID(sub->sid, service);
-        HandleUnlock();
         return;
     }
     /* set the timeout */
@@ -924,12 +841,9 @@ void gena_process_subscription_renewal_request(MHDTransaction *mhdt)
         sub->expireTime = time(nullptr) + time_out;
     }
 
-    if (respond_ok(mhdt, time_out, sub, handle_info->productversion)
-        != UPNP_E_SUCCESS) {
+    if (respond_ok(mhdt, time_out, sub, handle_info->productversion) != UPNP_E_SUCCESS) {
         RemoveSubscriptionSID(sub->sid, service);
     }
-
-    HandleUnlock();
 }
 
 
@@ -937,7 +851,7 @@ void gena_process_unsubscribe_request(MHDTransaction *mhdt)
 {
     UpnpPrintf(UPNP_DEBUG, GENA, __FILE__, __LINE__,
                    "gena_process_unsubscribe_request\n");
-    
+
     service_info *service;
     struct Handle_Info *handle_info;
     UpnpDevice_Handle device_handle;
@@ -956,26 +870,22 @@ void gena_process_unsubscribe_request(MHDTransaction *mhdt)
     }
     Upnp_SID sid = itsid->second;
 
-    HandleLock();
+    HANDLELOCK();
 
     if (GetDeviceHandleInfoForPath(
             mhdt->url, &device_handle, &handle_info, &service) != HND_DEVICE) {
         http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
-        HandleUnlock();
         return;
     }
 
     /* validate service */
     if (service == nullptr || !service->active || GetSubscriptionSID(sid, service) == nullptr) {
         http_SendStatusResponse(mhdt, HTTP_PRECONDITION_FAILED);
-        HandleUnlock();
         return;
     }
 
     RemoveSubscriptionSID(sid, service);
     http_SendStatusResponse(mhdt, HTTP_OK);
-
-    HandleUnlock();
 }
 
 #endif /* INCLUDE_DEVICE_APIS */
