@@ -36,19 +36,16 @@
 #include <cstdlib>
 #include <curl/curl.h>
 #include <iostream>
-#include <sstream>
 
+#include "genut.h"
 #include "miniserver.h"
-#include "httputils.h"
+#include "soaplib.h"
 #include "statcodes.h"
 #include "upnpapi.h"
-#include "soaplib.h"
 #include "uri.h"
-#include "upnp.h"
-#include "genut.h"
 
 #ifdef USE_EXPAT
-#include "expatmm.hxx"
+#include "expatmm.h"
 #define XMLPARSERTP inputRefXMLParser
 #else
 #include "picoxml.h"
@@ -153,6 +150,19 @@ get_response_value(
     return (*errcodep) ? SOAP_ACTION_RESP_ERROR : SOAP_ACTION_RESP;
 }
 
+class UPnPSoapOptParser {
+public:
+    UPnPSoapOptParser(const std::vector<std::pair<std::string, std::string>>& opts)
+    {
+        for (const auto& [name, val] : opts) {
+            if (name == "timeoutms") {
+                timeoutms = std::stoi(val);
+            }
+        }
+    }
+    long timeoutms{-1};
+};
+
 int SoapSendAction(
     const std::string& xml_header_str, const std::string& actionURL,
     const std::string& serviceType,    const std::string& actionName,
@@ -169,14 +179,16 @@ int SoapSendAction(
     const static std::string xml_body_start{"<s:Body>"};
     const static std::string xml_end{"</s:Body>\r\n</s:Envelope>\r\n"};
 
+    UPnPSoapOptParser opts(respdata);
+    respdata.clear();
+    long timeoutms = opts.timeoutms >= 0 ? opts.timeoutms : 1000L * HTTP_DEFAULT_TIMEOUT;
+    
     /* Action: name and namespace (servicetype) */
     std::ostringstream act;
     act << "<u:" << actionName << R"( xmlns:u=")" << serviceType << R"(">)" "\n";
     /* Action arguments */
-    for (const auto& arg : actionArgs) {
-        act << "<" << arg.first << ">" << xmlQuote(arg.second) << "</" <<
-            arg.first << ">\n";
-    }
+    for (const auto& [name, val] : actionArgs)
+        act << "<" << name << ">" << xmlQuote(val) << "</" << name << ">\n";
     act << "</u:" << actionName << ">\n";
     
     /* parse url */
@@ -214,7 +226,7 @@ int SoapSendAction(
         curl_easy_setopt(easy, CURLOPT_WRITEDATA, &responsestr);
         curl_easy_setopt(easy, CURLOPT_HEADERFUNCTION, header_callback_curl);
         curl_easy_setopt(easy, CURLOPT_HEADERDATA, &http_headers);
-        curl_easy_setopt(easy, CURLOPT_TIMEOUT, long(UPNP_TIMEOUT));
+        curl_easy_setopt(easy, CURLOPT_TIMEOUT_MS, timeoutms);
         curl_easy_setopt(easy, CURLOPT_POST, long(1));
         curl_easy_setopt(easy, CURLOPT_POSTFIELDS, payload.c_str()); 
         struct curl_slist *list = nullptr;
@@ -267,11 +279,10 @@ int SoapSendAction(
                "soapSendAction: http_stt [%ld] errcode %d errdesc[%s]\n",
                http_status, *errcodep, errdesc.c_str());
 
-    if (ret_code == SOAP_ACTION_RESP) {
+    if (ret_code == SOAP_ACTION_RESP)
         return UPNP_E_SUCCESS;
-    } else if (ret_code == SOAP_ACTION_RESP_ERROR ) {
+    if (ret_code == SOAP_ACTION_RESP_ERROR)
         return *errcodep;
-    }
     return ret_code;
 }
 
