@@ -39,23 +39,21 @@
 #include <cstring>
 
 #include <iostream>
-#include <sstream>
-#include <string>
 #include <map>
+#include <string>
 
 #ifdef USE_EXPAT
-#include "expatmm.hxx"
+#include "expatmm.h"
 #define XMLPARSERTP inputRefXMLParser
 #else
 #include "picoxml.h"
 #define XMLPARSERTP PicoXMLParser
 #endif
 
+#include "genut.h"
 #include "soaplib.h"
 #include "statcodes.h"
 #include "upnpapi.h"
-#include "httputils.h"
-#include "genut.h"
 
 #define SREQ_HDR_NOT_FOUND     -1
 #define SREQ_BAD_HDR_FORMAT     -2
@@ -119,8 +117,7 @@ static void send_error_response(
     const std::string& txt = ostr.str();
     mhdt->response = MHD_create_response_from_buffer(
         txt.size(), const_cast<char*>(txt.c_str()), MHD_RESPMEM_MUST_COPY);
-    MHD_add_response_header(mhdt->response, "Content-Type",
-                            R"(text/xml; charset="utf-8")");
+    MHD_add_response_header(mhdt->response, "Content-Type", R"(text/xml; charset="utf-8")");
     MHD_add_response_header(mhdt->response, "SERVER",
                             get_sdk_device_info(productversion).c_str());
     /* We do as the original code, but should this not be error_code? */
@@ -139,36 +136,27 @@ static void send_action_response(
     response << start_body;
     response << "<u:" << soap_info->action_name << "Response" <<
         R"( xmlns:u=")" << soap_info->service_type << R"(">)" "\n";
-    for (const auto&  arg : data) {
-        response << "<" << arg.first << ">" <<
-            xmlQuote(arg.second) <<
-            "</" <<    arg.first << ">\n";
-    }
+    for (const auto& [name, val] : data)
+        response << "<" << name << ">" << xmlQuote(val) << "</" << name << ">\n";
     response << "</u:" << soap_info->action_name << "Response" << ">\n";
     response << end_body;
     const std::string& txt(response.str());
-    UpnpPrintf(UPNP_INFO, SOAP, __FILE__, __LINE__,
-               "Action Response data: [%s]\n", txt.c_str());
+    UpnpPrintf(UPNP_INFO, SOAP, __FILE__, __LINE__, "Action Response data: [%s]\n", txt.c_str());
     mhdt->response = MHD_create_response_from_buffer(
         txt.size(), const_cast<char*>(txt.c_str()), MHD_RESPMEM_MUST_COPY);
     MHD_add_response_header(
-        mhdt->response, "SERVER",
-        get_sdk_device_info(soap_info->productversion).c_str());
+        mhdt->response, "SERVER", get_sdk_device_info(soap_info->productversion).c_str());
     mhdt->httpstatus = 200;
 }
 
 
 /* The original code performed a few consistency checks on the action xml
-   - Checked the soap namespace against
-     SOAP_URN = "http:/""/schemas.xmlsoap.org/soap/envelope/";
+   - Checked the soap namespace against SOAP_URN = "http:/""/schemas.xmlsoap.org/soap/envelope/";
    - Checked the "Body" elt name
-   - Checked that the action node namespace uri matched the service
-     type from the SOAPACTION header
-   - Checked that the action node local name matched the action name
-     from the SOAPACTION header.
+   - Checked that the action node namespace uri matched the service type from the SOAPACTION header
+   - Checked that the action node local name matched the action name from the SOAPACTION header.
    - Other checks for a var request. We don't support this any more.
-  As we're not in the business of checking conformity, we did not reproduce
-  the tests for now.
+  As we're not in the business of checking conformity, we did not reproduce the tests for now.
 */
 class UPnPActionRequestParser : public XMLPARSERTP {
 public:
@@ -232,7 +220,7 @@ static void handle_invoke_action(
 {
     struct Upnp_Action_Request action;
     int err_code;
-    const char *err_str;
+    const char *err_str = "";
 
     action.ErrCode = UPNP_E_SUCCESS;
     action.ErrStr[0] = 0;
@@ -244,8 +232,7 @@ static void handle_invoke_action(
     mhdt->copyClientAddress(&action.CtrlPtIPAddr);
     mhdt->copyHeader("user-agent", action.Os);
 
-    int ret = soap_info->callback(
-        UPNP_CONTROL_ACTION_REQUEST, &action, soap_info->cookie);
+    int ret = soap_info->callback(UPNP_CONTROL_ACTION_REQUEST, &action, soap_info->cookie);
     if (ret != UPNP_E_SUCCESS) {
         UpnpPrintf(UPNP_DEBUG, SOAP, __FILE__, __LINE__,
                    "Action callback failed. ret %d errcode %d errstr [%s]\n",
@@ -273,8 +260,7 @@ static void handle_invoke_action(
         UPnPActionRequestParser parser(action.xmlResponse, rspnm, args, true);
         if (!parser.Parse()) {
             UpnpPrintf(UPNP_INFO, SOAP, __FILE__, __LINE__,
-                       "XML response parse failed for [%s]\n",
-                       action.xmlResponse.c_str());
+                       "XML response parse failed for [%s]\n", action.xmlResponse.c_str());
             err_code = SOAP_ACTION_FAILED;
             err_str = Soap_Action_Failed;
             goto error_handler;
@@ -298,22 +284,18 @@ error_handler:
  *
  * \return 0 if OK, -1 on error.
  */
-static int get_dev_service(
-    MHDTransaction *mhdt, soap_devserv_t *soap_info)
+static int get_dev_service(const MHDTransaction *mhdt, soap_devserv_t *soap_info)
 {
     struct Handle_Info *hdlinfo;
     int device_hnd;
     service_info *serv_info;
 
-    HandleReadLock();
+    HANDLELOCK();
 
-    auto hdltp = GetDeviceHandleInfoForPath(
-        mhdt->url, &device_hnd, &hdlinfo, &serv_info);
+    auto hdltp = GetDeviceHandleInfoForPath(mhdt->url, &device_hnd, &hdlinfo, &serv_info);
 
     if (hdltp != HND_DEVICE || nullptr == serv_info) {
-        HandleUnlock();
-        UpnpPrintf(UPNP_ERROR, SOAP, __FILE__, __LINE__,
-                   "get_dev_service: client not found.\n");
+        UpnpPrintf(UPNP_ERROR, SOAP, __FILE__, __LINE__, "get_dev_service: client not found.\n");
         return -1;
     }
 
@@ -323,7 +305,6 @@ static int get_dev_service(
     soap_info->callback = hdlinfo->Callback;
     soap_info->cookie = hdlinfo->Cookie;
     soap_info->productversion = hdlinfo->productversion;
-    HandleUnlock();
     return 0;
 }
 
@@ -361,8 +342,7 @@ static int get_mpost_acton_hdrval(MHDTransaction *mhdt, std::string& val)
  *
  * \return UPNP_E_SUCCESS if OK, error number on failure.
  */
-static int check_soapaction_hdr(
-    MHDTransaction *mhdt, soap_devserv_t *soap_info)
+static int check_soapaction_hdr(MHDTransaction *mhdt, soap_devserv_t *soap_info)
 {
     int ret_code;
     std::string header;
@@ -383,24 +363,34 @@ static int check_soapaction_hdr(
     /* error by default */
     ret_code = SREQ_BAD_HDR_FORMAT;
 
-    /* The header value is something like: "urn:av-open...:Playlist:1#Id"  */
-    if (header[0] != '\"') {
-        return ret_code;
-    }
-
+    /* The header value is something like: "urn:av-open...:Playlist:1#Id" */
     std::string::size_type hash_pos = header.find('#');
     if (hash_pos == std::string::npos) {
         return ret_code;
     }
-    char anm[201];
-    if (sscanf(header.c_str() + hash_pos+1, "%200[^\"]", anm) != 1) {
-        return ret_code;
-    }
-    soap_info->action_name = anm;
 
-    /* check service type */
-    std::string serv_type = header.substr(1, hash_pos-1);
-    size_t cp1_diff = serv_type.rfind(':');
+    // Note: the UPnP standard says that this header value MUST be quoted with double quotes.
+    // Both npupnp and pupnp used to only check for a starting double quote, and accept a missing
+    // ending double quote.
+    // We are now even more lenient (following the pupnp lead, they made the same change), and
+    // accept an unquoted value,
+    std::string::size_type startadjust{1};
+    if (header[0] != '"') {
+        startadjust = 0;
+    }
+    std::string::size_type endadjust{1};
+    if (header.back() != '"') {
+        endadjust = 0;
+    }
+
+    // Action name: between the hash and the end (eos or double quote).
+    soap_info->action_name = header.substr(hash_pos + 1, header.size() - hash_pos - 1 - endadjust);
+
+    /* Service type: between start or double quote, and hash */
+    auto serv_type = header.substr(startadjust, hash_pos - startadjust);
+
+    /* Check service type */
+    auto cp1_diff = serv_type.rfind(':');
     if (std::string::npos == cp1_diff) {
         return ret_code;
     }
@@ -424,8 +414,6 @@ static int check_soapaction_hdr(
 }
 
 
-
-
 /*!
  * \brief This is a callback called by miniserver after receiving the request
  * from the control point. After HTTP processing, it calls handle_soap_request
@@ -438,7 +426,8 @@ void soap_device_callback(MHDTransaction *mhdt)
     soap_devserv_t soap_info;
     std::vector<std::pair<std::string, std::string>> args;
     std::string strippedxml;
-    std::string productversion;
+    UpnpPrintf(UPNP_INFO, SOAP, __FILE__, __LINE__, "Action POST data: [%s]\n",
+               mhdt->postdata.c_str());
     
     /* The device/service identified by the request URI */
     if (get_dev_service(mhdt, &soap_info) < 0) {
@@ -482,8 +471,7 @@ void soap_device_callback(MHDTransaction *mhdt)
         // header The parser will produce both argument vectors and an
         // XML subdocument matching the subtree which libupnp would
         // have sent (transition help).
-        UPnPActionRequestParser parser(
-            mhdt->postdata, soap_info.action_name, args, false);
+        UPnPActionRequestParser parser(mhdt->postdata, soap_info.action_name, args, false);
         if (!parser.Parse()) {
             UpnpPrintf(UPNP_INFO, SOAP, __FILE__, __LINE__,
                        "XML parse failed for [%s]\n", mhdt->postdata.c_str());
@@ -498,8 +486,7 @@ void soap_device_callback(MHDTransaction *mhdt)
     handle_invoke_action(mhdt, &soap_info, strippedxml, args);
 
     if (mhdt->response)
-        MHD_add_response_header(mhdt->response, "Content-Type",
-                                R"(text/xml; charset="utf-8")");
+        MHD_add_response_header(mhdt->response, "Content-Type", R"(text/xml; charset="utf-8")");
 
 error_handler:
     // productversion could be empty here, in which case we will send the lib
