@@ -2,7 +2,7 @@
  *
  * Copyright (c) 2000-2003 Intel Corporation
  * All rights reserved.
- * Copyright (C) 2011-2012 France Telecom All rights reserved. 
+ * Copyright (C) 2011-2012 France Telecom All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -35,30 +35,28 @@
 #ifdef INCLUDE_DEVICE_APIS
 #if EXCLUDE_SSDP == 0
 
+#include "ThreadPool.h"
+#include "genut.h"
 #include "httputils.h"
 #include "ssdplib.h"
 #include "statcodes.h"
-#include "ThreadPool.h"
-#include "upnpapi.h"
-#include "UpnpInet.h"
-#include "upnpdebug.h"
-#include "TimerThread.h"
-#include "genut.h"
-#include "inet_pton.h"
-#include "netif.h"
 #include "upnpapi.h"
 
-#include <cassert>
+#include <algorithm>
+#include <chrono>
 #include <cstdio>
 #include <cstring>
 #include <iostream>
 #include <sstream>
 #include <string>
 #include <thread>
-#include <algorithm>
-#include <chrono>
 
 struct SsdpSearchReply {
+    SsdpSearchReply(int a, UpnpDevice_Handle h, const sockaddr_storage* da, SsdpEntity e)
+        : MaxAge(a), handle(h), event(std::move(e))
+    {
+        std::memcpy(&dest_addr, da, sizeof(dest_addr));
+    }
     int MaxAge;
     UpnpDevice_Handle handle;
     struct sockaddr_storage dest_addr;
@@ -74,80 +72,68 @@ struct SSDPPwrState {
 // A bundle to simplify arg lists
 struct SSDPCommonData {
     SOCKET sock;
-    struct sockaddr *DestAddr;
-    const char *DevOrServType;
+    struct sockaddr_storage *DestAddr;
     SSDPPwrState pwr;
     std::string prodvers;
 };
 
-static void del_ssdpsearchreply(void *p)
-{
-    delete static_cast<SsdpSearchReply *>(p);
-}
+class SSDPSearchJobWorker : public JobWorker {
+public:
+    explicit SSDPSearchJobWorker(std::unique_ptr<SsdpSearchReply> reply)
+        : m_reply(std::move(reply)) {}
+    void work() override {
+        AdvertiseAndReply(m_reply->handle, MSGTYPE_REPLY,
+                          m_reply->MaxAge,
+                          &m_reply->dest_addr,
+                          m_reply->event);
+    }
+    std::unique_ptr<SsdpSearchReply> m_reply;
+};
 
-static void *thread_advertiseandreply(void *data)
-{
-    auto arg = static_cast<SsdpSearchReply *>(data);
-
-    AdvertiseAndReply(arg->handle, MSGTYPE_REPLY, 
-                      arg->MaxAge,
-                      reinterpret_cast<struct sockaddr *>(&arg->dest_addr),
-                      arg->event);
-    return nullptr;
-}
-
-void ssdp_handle_device_request(SSDPPacketParser& parser,
-                                struct sockaddr_storage *dest_addr)
+void ssdp_handle_device_request(const SSDPPacketParser& parser, struct sockaddr_storage *dest_addr)
 {
     int handle, start;
     struct Handle_Info *dev_info = nullptr;
-    int mx;
     SsdpEntity event;
-    int maxAge;
-
+    
     /* check man hdr. */
     if (!parser.man || strcmp(parser.man, R"("ssdp:discover")") != 0) {
         /* bad or missing hdr. */
-        UpnpPrintf(UPNP_ALL, API, __FILE__, __LINE__,
-                   "ssdp_handle_device_req: no/bad MAN header\n");
+        UpnpPrintf(UPNP_ALL, API, __FILE__, __LINE__, "ssdp_handle_device_req: no/bad MAN header\n");
         return;
     }
-    /* MX header. Must be >= 1*/
-    if (!parser.mx || (mx = atoi(parser.mx)) <= 0) {
-        UpnpPrintf(UPNP_ALL, API, __FILE__, __LINE__,
-                   "ssdp_handle_device_req: no/bad MX header\n");
-        return;
+    /* MX header. May be absent for a unicast request. Consistency has been checked in ssdp_server */
+    int mx = 0;
+    if (parser.mx) {
+        mx = atoi(parser.mx);
     }
     /* ST header. */
     if (!parser.st || ssdp_request_type(parser.st, &event) == -1) {
-        UpnpPrintf(UPNP_ALL, API, __FILE__, __LINE__,
-                   "ssdp_handle_device_req: no/bad ST header\n");
+        UpnpPrintf(UPNP_ALL, API, __FILE__, __LINE__, "ssdp_handle_device_req: no/bad ST header\n");
         return;
     }
 
-    // Loop to dispatch the packet to each of our configured devices
-    // by starting the handle search at the last processed
-    // position. each device response is scheduled by the ThreadPool
-    // with a random delay based on the MX header of the search packet.
+    // Loop to dispatch the packet to each of our configured devices by starting the handle search
+    // at the last processed position. each device response is scheduled by the ThreadPool with a
+    // random delay based on the MX header of the search packet.
     start = 0;
     for (;;) {
-        HandleLock();
-        /* device info. */
-        switch (GetDeviceHandleInfo(start, &handle, &dev_info)) {
-        case HND_DEVICE:
-            break;
-        default:
-            HandleUnlock();
-            /* no info found. */
-            return;
+        int maxAge;
+        {
+            HANDLELOCK();
+            /* device info. */
+            switch (GetDeviceHandleInfo(start, &handle, &dev_info)) {
+            case HND_DEVICE:
+                break;
+            default:
+                /* no info found. */
+                return;
+            }
+            maxAge = dev_info->MaxAge;
         }
-        maxAge = dev_info->MaxAge;
-        HandleUnlock();
 
-        UpnpPrintf(UPNP_DEBUG, API, __FILE__, __LINE__,
-                   "MAX-AGE        =  %d\n", maxAge);
-        UpnpPrintf(UPNP_DEBUG, API, __FILE__, __LINE__,
-                   "MX       =  %d\n", maxAge);
+        UpnpPrintf(UPNP_DEBUG, API, __FILE__, __LINE__, "MAX-AGE        =  %d\n", maxAge);
+        UpnpPrintf(UPNP_DEBUG, API, __FILE__, __LINE__, "MX       =  %d\n", mx);
         UpnpPrintf(UPNP_DEBUG, API, __FILE__, __LINE__,
                    "DeviceType     =    %s\n", event.DeviceType.c_str());
         UpnpPrintf(UPNP_DEBUG, API, __FILE__, __LINE__,
@@ -155,22 +141,19 @@ void ssdp_handle_device_request(SSDPPacketParser& parser,
         UpnpPrintf(UPNP_DEBUG, API, __FILE__, __LINE__,
                    "ServiceType =  %s\n", event.ServiceType.c_str());
 
-        auto threadArg = new SsdpSearchReply;
-        if (threadArg == nullptr)
-            return;
-        threadArg->handle = handle;
-        memcpy(&threadArg->dest_addr, dest_addr, sizeof(threadArg->dest_addr));
-        threadArg->event = event;
-        threadArg->MaxAge = maxAge;
-
-        mx = std::max(1, mx);
-        /* Subtract a bit from the mx to allow for network/processing delays */
-        int delayms = rand() %    (mx * 1000 - 100);
-        UpnpPrintf(UPNP_ALL, API, __FILE__, __LINE__,
-                   "ssdp_handle_device_req: scheduling resp in %d ms\n",delayms);
-        gTimerThread->schedule(
-            TimerThread::SHORT_TERM, std::chrono::milliseconds(delayms),
-            nullptr, thread_advertiseandreply, threadArg, del_ssdpsearchreply);
+        auto threadArg = std::make_unique<SsdpSearchReply>(maxAge, handle, dest_addr, event);
+        auto worker = std::make_unique<SSDPSearchJobWorker>(std::move(threadArg));
+        if (mx) {
+            mx = std::max(1, mx);
+            /* Subtract a bit from the mx to allow for network/processing delays */
+            int delayms = rand() % (mx * 1000 - 100);
+            UpnpPrintf(UPNP_ALL, API, __FILE__, __LINE__,
+                       "ssdp_handle_device_req: scheduling resp in %d ms\n", delayms);
+            gTimerThread->schedule(TimerThread::SHORT_TERM, std::chrono::milliseconds(delayms),
+                                   nullptr, std::move(worker));
+        } else {
+            gSendThreadPool.addJob(std::move(worker));
+        }
         start = handle;
     }
 }
@@ -263,7 +246,7 @@ static SOCKET createMulticastSocket6(int index, std::string& lochost)
             if (ipaddr) {
                 lochost = strInBrackets(ipaddr->straddr());
                 break;
-            } 
+            }
         }
     }
     if (lochost.empty()) {
@@ -276,7 +259,7 @@ error:
 }
 
 static SOCKET createReplySocket6(
-    struct sockaddr_in6 *destaddr, std::string& lochost)
+    const struct sockaddr_in6 *destaddr, std::string& lochost)
 {
     SOCKET sock = socket(AF_INET6, SOCK_DGRAM, 0);
     if (sock == INVALID_SOCKET) {
@@ -291,7 +274,7 @@ static SOCKET createReplySocket6(
             if (ipaddr) {
                 lochost = strInBrackets(ipaddr->straddr());
                 break;
-            } 
+            }
         }
     }
     if (lochost.empty()) {
@@ -334,34 +317,36 @@ static bool ssdpMcastAddr(struct sockaddr_storage& ss, int AddressFamily)
     return true;
 }
 
-static int sendPackets(SOCKET sock, struct sockaddr *daddr, int cnt, std::string *pckts)
+static int sendPackets(SOCKET sock, struct sockaddr_storage *daddr, int cnt, std::string *pckts)
 {
-    NetIF::IPAddr destip(daddr);
-    int socklen = daddr->sa_family == AF_INET ? sizeof(struct sockaddr_in) :
+    NetIF::IPAddr destip(reinterpret_cast<struct sockaddr*>(daddr));
+    int socklen = daddr->ss_family == AF_INET ? sizeof(struct sockaddr_in) :
         sizeof(struct sockaddr_in6);
 
     for (int i = 0; i < cnt; i++) {
         ssize_t rc;
         UpnpPrintf(UPNP_DEBUG, SSDP, __FILE__, __LINE__, ">>> SSDP SEND to %s >>>\n%s\n",
                    destip.straddr().c_str(), pckts[i].c_str());
-        rc = sendto(sock, pckts[i].c_str(), pckts[i].size(), 0, daddr, socklen);
-                    
+        rc = sendto(sock, pckts[i].c_str(), pckts[i].size(), 0,
+                    reinterpret_cast<struct sockaddr*>(daddr), socklen);
+
         if (rc == -1) {
-            char errorBuffer[ERROR_BUFFER_LEN];
-            posix_strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-            UpnpPrintf(UPNP_INFO,SSDP,__FILE__,__LINE__,"sendPackets: sendto: %s\n", errorBuffer);
+            std::string errorDesc;
+            NetIF::getLastError(errorDesc);
+            UpnpPrintf(UPNP_INFO,SSDP,__FILE__,__LINE__,
+                       "sendPackets: sendto: %s\n", errorDesc.c_str());
             return UPNP_E_SOCKET_WRITE;
         }
     }
     return UPNP_E_SUCCESS;
 }
 
- 
+
 /* Creates a device notify or search reply packet. */
 static void CreateServicePacket(
     SSDPDevMessageType msg_type, const char *nt, const char *usn,
     const std::string& location, int duration, std::string &packet,
-    int AddressFamily, SSDPPwrState& pwr, const std::string& prodvers)
+    int AddressFamily, const SSDPPwrState& pwr, const std::string& prodvers)
 {
     std::ostringstream str;
     switch (msg_type) {
@@ -422,13 +407,15 @@ static void CreateServicePacket(
             "SleepPeriod: " << pwr.SleepPeriod << "\r\n" <<
             "RegistrationState: " << pwr.RegistrationState << "\r\n";
     }
+    str << "BOOTID.UPNP.ORG: " << g_bootidUpnpOrg << "\r\n" <<
+        "CONFIGID.UPNP.ORG: " << g_configidUpnpOrg << "\r\n";
 
     str << "\r\n";
     packet = str.str();
 }
 
 static int DeviceAdvertisementOrShutdown(
-    SSDPCommonData& sscd, SSDPDevMessageType msgtype, const char *DevType,
+    const SSDPCommonData& sscd, SSDPDevMessageType msgtype, const char *DevType,
     int RootDev, const char *Udn, const std::string& Location, int Duration)
 {
     char Mil_Usn[LINE_SIZE];
@@ -443,17 +430,17 @@ static int DeviceAdvertisementOrShutdown(
             goto error_handler;
         CreateServicePacket(msgtype, "upnp:rootdevice",
                             Mil_Usn, Location, Duration, msgs[0],
-                            sscd.DestAddr->sa_family, sscd.pwr, sscd.prodvers);
+                            sscd.DestAddr->ss_family, sscd.pwr, sscd.prodvers);
     }
     /* both root and sub-devices need to send these two messages */
     CreateServicePacket(msgtype, Udn, Udn,
-                        Location, Duration, msgs[1], sscd.DestAddr->sa_family,
+                        Location, Duration, msgs[1], sscd.DestAddr->ss_family,
                         sscd.pwr, sscd.prodvers);
     rc = snprintf(Mil_Usn, sizeof(Mil_Usn), "%s::%s", Udn, DevType);
     if (rc < 0 || static_cast<unsigned int>(rc) >= sizeof(Mil_Usn))
         goto error_handler;
     CreateServicePacket(msgtype, DevType, Mil_Usn, Location, Duration, msgs[2],
-                        sscd.DestAddr->sa_family, sscd.pwr, sscd.prodvers);
+                        sscd.DestAddr->ss_family, sscd.pwr, sscd.prodvers);
     /* check error */
     if ((RootDev && msgs[0].empty()) || msgs[1].empty() || msgs[2].empty()) {
         goto error_handler;
@@ -472,7 +459,7 @@ error_handler:
 }
 
 static int SendReply(
-    SSDPCommonData& sscd, const char *DevType, int RootDev,
+    const SSDPCommonData& sscd, const char *DevType, int RootDev,
     const char *Udn, const std::string& Location, int Duration, int ByType)
 {
     int ret_code = UPNP_E_OUTOF_MEMORY;
@@ -481,7 +468,7 @@ static int SendReply(
     char Mil_Usn[LINE_SIZE];
     int i;
     int rc = 0;
-    int family = static_cast<int>(sscd.DestAddr->sa_family);
+    auto family = static_cast<int>(sscd.DestAddr->ss_family);
 
     if (RootDev) {
         /* one msg for root device */
@@ -522,14 +509,14 @@ error_handler:
 }
 
 static int DeviceReply(
-    SSDPCommonData& sscd, const char *DevType, int RootDev,
+    const SSDPCommonData& sscd, const char *DevType, int RootDev,
     const char *Udn, const std::string& Location, int Duration)
 {
     std::string szReq[3];
     char Mil_Nt[LINE_SIZE], Mil_Usn[LINE_SIZE];
     int rc = 0;
-    int family = static_cast<int>(sscd.DestAddr->sa_family);
-    
+    auto family = static_cast<int>(sscd.DestAddr->ss_family);
+
     /* create 2 or 3 msgs */
     if (RootDev) {
         /* 3 replies for root device */
@@ -568,7 +555,7 @@ static int DeviceReply(
 }
 
 static int ServiceSend(
-    SSDPCommonData& sscd, SSDPDevMessageType tp, const char *ServType,
+    const SSDPCommonData& sscd, SSDPDevMessageType tp, const char *ServType,
     const char *Udn, const std::string& Location, int Duration)
 {
     char Mil_Usn[LINE_SIZE];
@@ -579,12 +566,12 @@ static int ServiceSend(
     if (rc < 0 || static_cast<unsigned int>(rc) >= sizeof(Mil_Usn))
         return UPNP_E_OUTOF_MEMORY;
     CreateServicePacket(tp, ServType, Mil_Usn, Location, Duration, szReq[0],
-                        sscd.DestAddr->sa_family, sscd.pwr, sscd.prodvers);
+                        sscd.DestAddr->ss_family, sscd.pwr, sscd.prodvers);
     if (szReq[0].empty()) {
         return UPNP_E_OUTOF_MEMORY;
 
     }
-    
+
     return sendPackets(sscd.sock, sscd.DestAddr, 1, szReq);
 }
 
@@ -623,42 +610,48 @@ static bool sameServOrDevNoVers(const char *his, const char *mine)
 // be subdevices
 static int AdvertiseAndReplyOneDest(
     UpnpDevice_Handle Hnd, SSDPDevMessageType tp, int Exp,
-    struct sockaddr *DestAddr, const SsdpEntity& sdata, SOCKET sock, 
+    struct sockaddr_storage *DestAddr, const SsdpEntity& sdata, SOCKET sock,
     const std::string& lochost)
 {
     int retVal = UPNP_E_SUCCESS;
     int NumCopy = 0;
-    std::vector<const UPnPDeviceDesc*> alldevices;
+    std::vector<UPnPDeviceDesc> alldevices;
     bool isNotify = (tp == MSGTYPE_ADVERTISEMENT || tp == MSGTYPE_SHUTDOWN);
+    struct Handle_Info *SInfoPtr;
+    struct Handle_Info SInfo;
+    std::string location, lowerloc;
     
-    /* Use a read lock */
-    HandleReadLock();
-    struct Handle_Info *SInfo = nullptr;
-    if (GetHandleInfo(Hnd, &SInfo) != HND_DEVICE) {
-        HandleUnlock();
-        return UPNP_E_INVALID_HANDLE;
+    {
+        // We make copies of everything with the handle table lock held, so as not to hold it while
+        // sending the advertisements. Not sure that this is actually useful.
+        HANDLELOCK();
+        if (GetHandleInfo(Hnd, &SInfoPtr) != HND_DEVICE) {
+            return UPNP_E_INVALID_HANDLE;
+        }
+        SInfo.MaxAge = SInfoPtr->MaxAge;
+        SInfo.PowerState = SInfoPtr->PowerState;
+        SInfo.SleepPeriod = SInfoPtr->SleepPeriod;
+        SInfo.RegistrationState = SInfoPtr->RegistrationState;
+        SInfo.productversion = SInfoPtr->productversion;
+        memcpy(SInfo.DescURL, SInfoPtr->DescURL, LINE_SIZE);
+        memcpy(SInfo.LowerDescURL, SInfoPtr->LowerDescURL, LINE_SIZE);
+        // Store pointers to the root and embedded devices in a single vector
+        // for later convenience of mostly identical processing.
+        alldevices.push_back(SInfoPtr->devdesc);
+        for (const auto& dev : SInfoPtr->devdesc.embedded) {
+            alldevices.push_back(dev);
+        }
+        location = SInfoPtr->DescURL;
+        replaceLochost(location, lochost);
+        lowerloc = SInfoPtr->LowerDescURL;
+        replaceLochost(lowerloc, lochost);
     }
 
-    int defaultExp = SInfo->MaxAge;
+    int defaultExp = SInfo.MaxAge;
+    SSDPCommonData sscd{sock, DestAddr,
+                        SSDPPwrState{SInfo.PowerState, SInfo.SleepPeriod,
+                                     SInfo.RegistrationState}, SInfo.productversion};
 
-    struct SSDPCommonData sscd;
-    sscd.sock = sock;
-    sscd.DestAddr = DestAddr;
-    sscd.pwr = SSDPPwrState{SInfo->PowerState, SInfo->SleepPeriod, SInfo->RegistrationState};
-    sscd.prodvers = SInfo->productversion;
-
-    
-    std::string location{SInfo->DescURL};
-    replaceLochost(location, lochost);
-    std::string lowerloc{SInfo->LowerDescURL};
-    replaceLochost(lowerloc, lochost);
-
-    // Store pointers to the root and embedded devices in a single vector
-    // for later convenience of mostly identical processing.
-    alldevices.push_back(&SInfo->devdesc);
-    for (const auto& dev : SInfo->devdesc.embedded) {
-        alldevices.push_back(&dev);
-    }
 
     /* send advertisements/replies */
     while (NumCopy == 0 || (isNotify && NumCopy < NUM_SSDP_COPY)) {
@@ -668,8 +661,8 @@ static int AdvertiseAndReplyOneDest(
 
         for (auto& devp : alldevices) {
             bool isroot = &devp == &(*alldevices.begin());
-            const char *devType = devp->deviceType.c_str();
-            const char *UDNstr = devp->UDN.c_str();
+            const char *devType = devp.deviceType.c_str();
+            const char *UDNstr = devp.UDN.c_str();
             if (isNotify) {
                 DeviceAdvertisementOrShutdown(sscd, tp, devType,  isroot, UDNstr, location, Exp);
             } else {
@@ -693,7 +686,7 @@ static int AdvertiseAndReplyOneDest(
                         UpnpPrintf(UPNP_DEBUG, SSDP, __FILE__, __LINE__,
                                    "DeviceUDN=%s/search UDN=%s NOMATCH\n",
                                    UDNstr, sdata.UDN.c_str());
-                    }                            
+                    }
                     break;
 
                 case SSDP_DEVICETYPE: {
@@ -736,7 +729,7 @@ static int AdvertiseAndReplyOneDest(
              * is directly traversed as a child of its parent device. This
              * ensures that the service's alive message uses the UDN of
              * the parent device. */
-            for (const auto& service : devp->services) {
+            for (const auto& service : devp.services) {
                 const char *servType = service.serviceType.c_str();
                 UpnpPrintf(UPNP_DEBUG, SSDP, __FILE__, __LINE__, "ServiceType = %s\n", servType);
                 if (isNotify) {
@@ -787,7 +780,6 @@ static int AdvertiseAndReplyOneDest(
     }
 
     UpnpPrintf(UPNP_ALL, SSDP, __FILE__, __LINE__, "AdvertiseAndReply1 exit\n");
-    HandleUnlock();
     return retVal;
 }
 
@@ -799,7 +791,7 @@ static int AdvertiseAndReplyOneDest(
 // work for each of the appropriate network addresses/interfaces.
 
 int AdvertiseAndReply(UpnpDevice_Handle Hnd, SSDPDevMessageType tp, int Exp,
-                      struct sockaddr *repDestAddr, const SsdpEntity& sdata)
+                      struct sockaddr_storage *repDestAddr, const SsdpEntity& sdata)
 {
     bool isNotify = (tp == MSGTYPE_ADVERTISEMENT || tp == MSGTYPE_SHUTDOWN);
     int ret = UPNP_E_SUCCESS;
@@ -813,7 +805,7 @@ int AdvertiseAndReply(UpnpDevice_Handle Hnd, SSDPDevMessageType tp, int Exp,
                        "ssdp_device: mcast for %s\n", netif.getname().c_str());
 
             struct sockaddr_storage dss;
-            auto destaddr = reinterpret_cast<struct sockaddr*>(&dss);
+            auto destaddr = &dss;
 
 #ifdef UPNP_ENABLE_IPV6
             if (using_ipv6()) {
@@ -825,7 +817,7 @@ int AdvertiseAndReply(UpnpDevice_Handle Hnd, SSDPDevMessageType tp, int Exp,
 
                 ret = AdvertiseAndReplyOneDest(
                     Hnd, tp, Exp, destaddr, sdata, sock, lochost);
-                
+
                 if (ret != UPNP_E_SUCCESS) {
                     UpnpPrintf(UPNP_INFO, SSDP, __FILE__, __LINE__,
                                "SSDP dev: IPV6 SEND failed for %s\n", netif.getname().c_str());
@@ -837,16 +829,13 @@ int AdvertiseAndReply(UpnpDevice_Handle Hnd, SSDPDevMessageType tp, int Exp,
 #endif /* UPNP_ENABLE_IPV6 */
 
             ssdpMcastAddr(dss, AF_INET);
-            auto addresses = netif.getaddresses();
-            for (const auto& ipaddr : addresses.first) {
-                if (ipaddr.family() == NetIF::IPAddr::Family::IPV4) {
-                    const struct sockaddr_storage& fss{ipaddr.getaddr()};
-                    sock = createMulticastSocket4(
-                        reinterpret_cast<const struct sockaddr_in*>(&fss),
-                        lochost);
-                } else {
+            for (const auto& ipaddr : netif.getaddresses().first) {
+                if (ipaddr.family() != NetIF::IPAddr::Family::IPV4)
                     continue;
-                }
+                const struct sockaddr_storage& fss{ipaddr.getaddr()};
+                sock = createMulticastSocket4(
+                    reinterpret_cast<const struct sockaddr_in*>(&fss),
+                    lochost);
                 if (sock == INVALID_SOCKET) {
                     goto exitfunc;
                 }
@@ -858,7 +847,7 @@ int AdvertiseAndReply(UpnpDevice_Handle Hnd, SSDPDevMessageType tp, int Exp,
             }
         }
     } else {
-        sock = repDestAddr->sa_family == AF_INET ?
+        sock = repDestAddr->ss_family == AF_INET ?
             createReplySocket4(
                 reinterpret_cast<struct sockaddr_in*>(repDestAddr), lochost) :
             createReplySocket6(
@@ -867,18 +856,17 @@ int AdvertiseAndReply(UpnpDevice_Handle Hnd, SSDPDevMessageType tp, int Exp,
             goto exitfunc;
         }
         ret = AdvertiseAndReplyOneDest(
-            Hnd, tp, Exp, repDestAddr, sdata, sock, lochost);            
+            Hnd, tp, Exp, repDestAddr, sdata, sock, lochost);
     }
 
 exitfunc:
     if (ret != UPNP_E_SUCCESS) {
-        char errorBuffer[ERROR_BUFFER_LEN];
-        posix_strerror_r(errno, errorBuffer, ERROR_BUFFER_LEN);
-        UpnpPrintf(UPNP_INFO, SSDP, __FILE__, __LINE__,
-                   "sendPackets: %s\n", errorBuffer);
+        std::string errorDesc;
+        NetIF::getLastError(errorDesc);
+        UpnpPrintf(UPNP_INFO, SSDP, __FILE__, __LINE__, "sendPackets: %s\n", errorDesc.c_str());
         return UPNP_E_NETWORK_ERROR;
     }
-    if (sock != INVALID_SOCKET) 
+    if (sock != INVALID_SOCKET)
         UpnpCloseSocket(sock);
     return ret;
 }
